#! /usr/bin/env python3
#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#

import argparse
from dataclasses import dataclass
from enum import StrEnum
import os
from pathlib import Path
import re
import sys
from typing import *

import rich
from rich_argparse import RichHelpFormatter

import klayout.db as kdb
import klayout.rdb as rdb

from klayout_pex.log import (
    LogLevel,
    set_log_level,
    register_additional_handler,
    deregister_additional_handler,
    # console,
    # debug,
    info,
    warning,
    subproc,
    error,
    rule
)

from klayout_pex.magic.magic_ext_file_parser import (
    parse_magic_ext_file,
    parse_magic_res_ext_file
)


PROGRAM_NAME = "magic_log_analyzer"


class MagicLogAnalyzer:
    def __init__(self,
                 magic_log_dir_path: str,
                 report: rdb.ReportDatabase,
                 dbu: float):
        self.magic_log_dir_path = Path(magic_log_dir_path)
        self.report = report
        self.magic_category = self.report.create_category('MAGIC Extraction')
        self.dbu = dbu

    def analyze(self):
        # search for <cell>.ext files (C related)
        # search for <cell>.res.ext files (R related)
        # ext_files = [f.resolve() for f in self.magic_log_dir_path.glob('*.ext')]
        ext_files = self.magic_log_dir_path.glob('*.ext')

        CellName = str
        main_paths_by_cell_name: Dict[CellName, Path] = dict()
        res_paths_by_cell_name: Dict[CellName, Path] = dict()

        regex = r'(?P<cell>.*?)(?P<res>\.res)?\.ext'
        for ef in ext_files:
            m = re.match(regex, ef.name)
            if not m:
                continue

            cell = m.group('cell')
            res = m.group('res')

            if res:
                res_paths_by_cell_name[cell] = ef
            else:
                main_paths_by_cell_name[cell] = ef

        if not main_paths_by_cell_name:
            raise Exception(f"Could not find any *.ext files to analyze in {self.magic_log_dir_path}")

        for cell, ext_path in main_paths_by_cell_name.items():
            self.analyze_cell(cell=cell,
                              main_ext_path=ext_path,
                              res_ext_path=res_paths_by_cell_name.get(cell, None))

    def analyze_cell(self,
                     cell: str,
                     main_ext_path: Path,
                     res_ext_path: Optional[Path]):
        rdb_cell = self.report.create_cell(name=cell)
        ports_cat = self.report.create_category(parent=self.magic_category, name='Ports')
        nodes_cat = self.report.create_category(parent=self.magic_category, name='Nodes')
        devices_cat = self.report.create_category(parent=self.magic_category, name='Devices')
        rnodes_cat = self.report.create_category(parent=self.magic_category, name='Resistor Nodes')
        resistors_cat = self.report.create_category(parent=self.magic_category, name='Resistors')

        ext_data = parse_magic_ext_file(main_ext_path)
        res_ext_data = None if res_ext_path is None else parse_magic_res_ext_file(res_ext_path)

        dbu_to_um = 200.0

        def box_for_point_dbu(x: float, y: float) -> kdb.Box:
            return kdb.Box(x, y, x + 20, y + 20)

        for p in ext_data.ports:
            port_cat = self.report.create_category(parent=ports_cat, name=f"{p.net} ({p.layer})")
            shapes = kdb.Shapes()
            shapes.insert(kdb.Box(p.x_bot / dbu_to_um / self.dbu,
                                  p.y_bot / dbu_to_um / self.dbu,
                                  p.x_top / dbu_to_um / self.dbu,
                                  p.y_top / dbu_to_um / self.dbu))
            self.report.create_items(cell_id=rdb_cell.rdb_id(), category_id=port_cat.rdb_id(),
                                     trans=kdb.CplxTrans(mag=self.dbu), shapes=shapes)

        for n in ext_data.nodes:
            node_cat = self.report.create_category(parent=nodes_cat, name=f"{n.net} ({n.layer})")
            shapes = kdb.Shapes()
            shapes.insert(box_for_point_dbu(n.x_bot / dbu_to_um / self.dbu,
                                            n.y_bot / dbu_to_um / self.dbu))
            self.report.create_items(cell_id=rdb_cell.rdb_id(), category_id=node_cat.rdb_id(),
                                     trans=kdb.CplxTrans(mag=self.dbu), shapes=shapes)

        for d in ext_data.devices:
            device_cat = self.report.create_category(parent=devices_cat,
                                                     name=f"Type={d.device_type} Model={d.model}")
            shapes = kdb.Shapes()
            shapes.insert(kdb.Box(d.x_bot / dbu_to_um / self.dbu,
                                  d.y_bot / dbu_to_um / self.dbu,
                                  d.x_top / dbu_to_um / self.dbu,
                                  d.y_top / dbu_to_um / self.dbu))
            self.report.create_items(cell_id=rdb_cell.rdb_id(), category_id=device_cat.rdb_id(),
                                     trans=kdb.CplxTrans(mag=self.dbu), shapes=shapes)

        if res_ext_data is not None:
            for n in res_ext_data.rnodes:
                rnode_cat = self.report.create_category(parent=rnodes_cat,
                                                        name=n.name)
                shapes = kdb.Shapes()
                shapes.insert(box_for_point_dbu(n.x_bot / dbu_to_um / self.dbu,
                                                n.y_bot / dbu_to_um / self.dbu))
                self.report.create_items(cell_id=rdb_cell.rdb_id(), category_id=rnode_cat.rdb_id(),
                                         trans=kdb.CplxTrans(mag=self.dbu), shapes=shapes)

            for idx, r in enumerate(res_ext_data.resistors):
                res_cat = self.report.create_category(parent=resistors_cat,
                                                      name=f"#{idx} {r.node1}↔︎{r.node2} = {r.value_ohm} Ω")
                shapes = kdb.Shapes()
                for n in res_ext_data.rnodes_by_name(r.node1) + \
                         res_ext_data.rnodes_by_name(r.node2):
                    box = box_for_point_dbu(n.x_bot / dbu_to_um / self.dbu,
                                            n.y_bot / dbu_to_um / self.dbu)
                    shapes.insert(box)
                self.report.create_items(cell_id=rdb_cell.rdb_id(), category_id=res_cat.rdb_id(),
                                         trans=kdb.CplxTrans(mag=self.dbu), shapes=shapes)


class ArgumentValidationError(Exception):
    pass


def _parse_args(arg_list: List[str] = None) -> argparse.Namespace:
    main_parser = argparse.ArgumentParser(description=f"{PROGRAM_NAME}: "
                                                      f"Tool to create KLayout RDB for magic runs",
                                          add_help=False,
                                          formatter_class=RichHelpFormatter)

    main_parser.add_argument("--magic_log_dir", "-m",
                             dest="magic_log_dir_path", required=True,
                             help="Input magic log directory path")

    main_parser.add_argument("--out", "-o",
                             dest="output_rdb_path", default=None,
                             help="Magic log directory path (default is input directory / 'report.rdb.gz')")

    if arg_list is None:
        arg_list = sys.argv[1:]
    args = main_parser.parse_args(arg_list)

    if not os.path.isdir(args.magic_log_dir_path):
        raise ArgumentValidationError(f"Intput magic log directory does not exist at '{args.magic_log_dir_path}'")

    if args.output_rdb_path is None:
        os.path.join(args.magic_log_dir_path, 'report.rdb.gz')

    return args


def main():
    args = _parse_args()
    report = rdb.ReportDatabase('')
    c = MagicLogAnalyzer(magic_log_dir_path=args.magic_log_dir_path,
                         report=report,
                         dbu=1e-3)
    c.analyze()
    report.save(args.output_rdb_path)


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX 
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#

from __future__ import annotations  # allow class type hints within same class
from typing import *
from functools import cached_property
import google.protobuf.json_format

from .util.multiple_choice import MultipleChoicePattern
import klayout_pex_protobuf.tech_pb2 as tech_pb2
import klayout_pex_protobuf.process_stack_pb2 as process_stack_pb2
import klayout_pex_protobuf.process_parasitics_pb2 as process_parasitics_pb2

class TechInfo:
    """Helper class for Protocol Buffer tech_pb2.Technology"""

    GDSPair = Tuple[int, int]

    @staticmethod
    def parse_tech_def(jsonpb_path: str) -> tech_pb2.Technology:
        with open(jsonpb_path, 'r') as f:
            contents = f.read()
            tech = google.protobuf.json_format.Parse(contents, tech_pb2.Technology())
            return tech

    @classmethod
    def from_json(cls,
                  jsonpb_path: str,
                  dielectric_filter: Optional[MultipleChoicePattern]) -> TechInfo:
        tech = cls.parse_tech_def(jsonpb_path=jsonpb_path)
        return TechInfo(tech=tech,
                        dielectric_filter=dielectric_filter)

    def __init__(self,
                 tech: tech_pb2.Technology,
                 dielectric_filter: Optional[MultipleChoicePattern]):
        self.tech = tech
        self.dielectric_filter = dielectric_filter or MultipleChoicePattern(pattern='all')

    @cached_property
    def gds_pair_for_computed_layer_name(self) -> Dict[str, GDSPair]:
        return {lyr.layer_info.name: (lyr.layer_info.gds_layer, lyr.layer_info.gds_datatype)
                for lyr in self.tech.lvs_computed_layers}

    @cached_property
    def computed_layer_info_by_name(self) -> Dict[str, tech_pb2.ComputedLayerInfo]:
        return {lyr.layer_info.name: lyr for lyr in self.tech.lvs_computed_layers}

    @cached_property
    def computed_layer_info_by_gds_pair(self) -> Dict[GDSPair, tech_pb2.ComputedLayerInfo]:
        return {
            (lyr.layer_info.gds_layer, lyr.layer_info.gds_datatype): lyr
            for lyr in self.tech.lvs_computed_layers
        }

    @cached_property
    def canonical_layer_name_by_gds_pair(self) -> Dict[GDSPair, str]:
        return {
            (lyr.layer_info.gds_layer, lyr.layer_info.gds_datatype): lyr.original_layer_name
            for lyr in self.tech.lvs_computed_layers
        }

    @cached_property
    def layer_info_by_name(self) -> Dict[str, tech_pb2.LayerInfo]:
        return {lyr.name: lyr for lyr in self.tech.layers}

    @cached_property
    def gds_pair_for_layer_name(self) -> Dict[str, GDSPair]:
        return {lyr.name: (lyr.gds_layer, lyr.gds_datatype) for lyr in self.tech.layers}

    @cached_property
    def layer_info_by_gds_pair(self) -> Dict[GDSPair, tech_pb2.LayerInfo]:
        return {(lyr.gds_layer, lyr.gds_datatype): lyr for lyr in self.tech.layers}

    @cached_property
    def process_stack_layer_by_name(self) -> Dict[str, process_stack_pb2.ProcessStackInfo.LayerInfo]:
        return {lyr.name: lyr for lyr in self.tech.process_stack.layers}

    @cached_property
    def process_stack_layer_by_gds_pair(self) -> Dict[GDSPair, process_stack_pb2.ProcessStackInfo.LayerInfo]:
        return {
            (lyr.gds_layer, lyr.gds_datatype): self.process_stack_layer_by_name[lyr.name]
            for lyr in self.tech.process_stack.layers
        }

    @cached_property
    def process_substrate_layer(self) -> process_stack_pb2.ProcessStackInfo.LayerInfo:
        return list(
            filter(lambda lyr: lyr.layer_type is process_stack_pb2.ProcessStackInfo.LAYER_TYPE_SUBSTRATE,
                   self.tech.process_stack.layers)
        )[0]

    @cached_property
    def process_diffusion_layers(self) -> List[process_stack_pb2.ProcessStackInfo.LayerInfo]:
        return list(
            filter(lambda lyr: lyr.layer_type is process_stack_pb2.ProcessStackInfo.LAYER_TYPE_DIFFUSION,
                   self.tech.process_stack.layers)
        )

    @cached_property
    def gate_poly_layer(self) -> process_stack_pb2.ProcessStackInfo.LayerInfo:
        return self.process_metal_layers[0]

    @cached_property
    def field_oxide_layer(self) -> process_stack_pb2.ProcessStackInfo.LayerInfo:
        return list(
            filter(lambda lyr: lyr.layer_type is process_stack_pb2.ProcessStackInfo.LAYER_TYPE_FIELD_OXIDE,
                   self.tech.process_stack.layers)
        )[0]

    @cached_property
    def process_metal_layers(self) -> List[process_stack_pb2.ProcessStackInfo.LayerInfo]:
        return list(
            filter(lambda lyr: lyr.layer_type == process_stack_pb2.ProcessStackInfo.LAYER_TYPE_METAL,
                   self.tech.process_stack.layers)
        )

    @cached_property
    def filtered_dielectric_layers(self) -> List[process_stack_pb2.ProcessStackInfo.LayerInfo]:
        layers = []
        for pl in self.tech.process_stack.layers:
            match pl.layer_type:
                case process_stack_pb2.ProcessStackInfo.LAYER_TYPE_SIMPLE_DIELECTRIC | \
                     process_stack_pb2.ProcessStackInfo.LAYER_TYPE_CONFORMAL_DIELECTRIC | \
                     process_stack_pb2.ProcessStackInfo.LAYER_TYPE_SIDEWALL_DIELECTRIC:
                    if self.dielectric_filter.is_included(pl.name):
                        layers.append(pl)
        return layers

    @cached_property
    def dielectric_by_name(self) -> Dict[str, float]:
        diel_by_name = {}
        for pl in self.filtered_dielectric_layers:
            match pl.layer_type:
                case process_stack_pb2.ProcessStackInfo.LAYER_TYPE_SIMPLE_DIELECTRIC:
                    diel_by_name[pl.name] = pl.simple_dielectric_layer.dielectric_k
                case process_stack_pb2.ProcessStackInfo.LAYER_TYPE_CONFORMAL_DIELECTRIC:
                    diel_by_name[pl.name] = pl.conformal_dielectric_layer.dielectric_k
                case process_stack_pb2.ProcessStackInfo.LAYER_TYPE_SIDEWALL_DIELECTRIC:
                    diel_by_name[pl.name] = pl.sidewall_dielectric_layer.dielectric_k
        return diel_by_name

    def sidewall_dielectric_layer(self, layer_name: str) -> Optional[process_stack_pb2.ProcessStackInfo.LayerInfo]:
        found_layers: List[process_stack_pb2.ProcessStackInfo.LayerInfo] = []
        for lyr in self.filtered_dielectric_layers:
            match lyr.layer_type:
                case process_stack_pb2.ProcessStackInfo.LAYER_TYPE_SIDEWALL_DIELECTRIC:
                    if lyr.sidewall_dielectric_layer.reference == layer_name:
                        found_layers.append(lyr)
                case process_stack_pb2.ProcessStackInfo.LAYER_TYPE_CONFORMAL_DIELECTRIC:
                    if lyr.conformal_dielectric_layer.reference == layer_name:
                        found_layers.append(lyr)
                case _:
                    continue

        if len(found_layers) == 0:
            return None
        if len(found_layers) >= 2:
            raise Exception(f"found multiple sidewall dielectric layers for {layer_name}")
        return found_layers[0]

    def simple_dielectric_above_metal(self, layer_name: str) -> Tuple[Optional[process_stack_pb2.ProcessStackInfo.LayerInfo], float]:
        """
        Returns a tuple of the dielectric layer and it's (maximum) height.
        Maximum would be the case where no metal and other dielectrics are present.
        """
        found_layer: Optional[process_stack_pb2.ProcessStackInfo.LayerInfo] = None
        diel_lyr: Optional[process_stack_pb2.ProcessStackInfo.LayerInfo] = None
        for lyr in self.tech.process_stack.layers:
            if lyr.name == layer_name:
                found_layer = lyr
            elif found_layer:
                if not diel_lyr and lyr.layer_type == process_stack_pb2.ProcessStackInfo.LAYER_TYPE_SIMPLE_DIELECTRIC:
                    if not self.dielectric_filter.is_included(lyr.name):
                        return None, 0.0
                    diel_lyr = lyr
                # search for next metal or end of stack
                if lyr.layer_type == process_stack_pb2.ProcessStackInfo.LAYER_TYPE_METAL:
                    return diel_lyr, lyr.metal_layer.height - found_layer.metal_layer.height
        return diel_lyr, 5.0   # air TODO

    @cached_property
    def substrate_cap_by_layer_name(self) -> Dict[str, process_parasitics_pb2.CapacitanceInfo.SubstrateCapacitance]:
        return {sc.layer_name: sc for sc in self.tech.process_parasitics.capacitance.substrates}

    @cached_property
    def overlap_cap_by_layer_names(self) -> Dict[str, Dict[str, process_parasitics_pb2.CapacitanceInfo.OverlapCapacitance]]:
        """
        usage: dict[top_layer_name][bottom_layer_name]
        """

        def convert_substrate_to_overlap_cap(sc: process_parasitics_pb2.CapacitanceInfo.SubstrateCapacitance) \
            -> process_parasitics_pb2.CapacitanceInfo.OverlapCapacitance:
            oc = process_parasitics_pb2.CapacitanceInfo.OverlapCapacitance()
            oc.top_layer_name = sc.layer_name
            oc.bottom_layer_name = self.internal_substrate_layer_name
            oc.capacitance = sc.area_capacitance
            return oc

        d = {
            ln: {
                self.internal_substrate_layer_name: convert_substrate_to_overlap_cap(sc)
            } for ln, sc in self.substrate_cap_by_layer_name.items()
        }

        d2 = {
            oc.top_layer_name: {
                oc_bot.bottom_layer_name: oc_bot
                for oc_bot in self.tech.process_parasitics.capacitance.overlaps if oc_bot.top_layer_name == oc.top_layer_name
            }
            for oc in self.tech.process_parasitics.capacitance.overlaps
        }

        for k1, ve in d2.items():
            for k2, v in ve.items():
                if k1 not in d:
                    d[k1] = {k2: v}
                else:
                    d[k1][k2] = v
        return d

    @cached_property
    def sidewall_cap_by_layer_name(self) -> Dict[str, process_parasitics_pb2.CapacitanceInfo.SidewallCapacitance]:
        return {sc.layer_name: sc for sc in self.tech.process_parasitics.capacitance.sidewalls}

    @classmethod
    @property
    def internal_substrate_layer_name(cls) -> str:
        return 'VSUBS'

    @cached_property
    def side_overlap_cap_by_layer_names(self) -> Dict[str, Dict[str, process_parasitics_pb2.CapacitanceInfo.SideOverlapCapacitance]]:
        """
        usage: dict[in_layer_name][out_layer_name]
        """

        def convert_substrate_to_side_overlap_cap(sc: process_parasitics_pb2.CapacitanceInfo.SubstrateCapacitance) \
            -> process_parasitics_pb2.CapacitanceInfo.SideOverlapCapacitance:
            soc = process_parasitics_pb2.CapacitanceInfo.SideOverlapCapacitance()
            soc.in_layer_name = sc.layer_name
            soc.out_layer_name = self.internal_substrate_layer_name
            soc.capacitance = sc.perimeter_capacitance
            return soc

        d = {
            ln: {
                self.internal_substrate_layer_name: convert_substrate_to_side_overlap_cap(sc)
            } for ln, sc in self.substrate_cap_by_layer_name.items()
        }

        d2 = {
            oc.in_layer_name: {
                oc_bot.out_layer_name: oc_bot
                for oc_bot in self.tech.process_parasitics.capacitance.sideoverlaps if oc_bot.in_layer_name == oc.in_layer_name
            }
            for oc in self.tech.process_parasitics.capacitance.sideoverlaps
        }

        for k1, ve in d2.items():
            for k2, v in ve.items():
                d[k1][k2] = v

        return d



#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX 
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#
__version__ = "0.1.15"



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: process_parasitics.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'process_parasitics.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18process_parasitics.proto\x12\tkpex.tech\"\x8a\x01\n\x15ProcessParasiticsInfo\x12\x11\n\tside_halo\x18\n \x01(\x05\x12-\n\nresistance\x18n \x01(\x0b\x32\x19.kpex.tech.ResistanceInfo\x12/\n\x0b\x63\x61pacitance\x18o \x01(\x0b\x32\x1a.kpex.tech.CapacitanceInfo\"\x98\x02\n\x0eResistanceInfo\x12\x39\n\x06layers\x18\x01 \x03(\x0b\x32).kpex.tech.ResistanceInfo.LayerResistance\x12\x35\n\x04vias\x18\x02 \x03(\x0b\x32\'.kpex.tech.ResistanceInfo.ViaResistance\x1a]\n\x0fLayerResistance\x12\x12\n\nlayer_name\x18\x01 \x01(\t\x12\x12\n\nresistance\x18\x02 \x01(\x01\x12\"\n\x1a\x63orner_adjustment_fraction\x18\x03 \x01(\x01\x1a\x35\n\rViaResistance\x12\x10\n\x08via_name\x18\x01 \x01(\t\x12\x12\n\nresistance\x18\x02 \x01(\x01\"\x98\x05\n\x0f\x43\x61pacitanceInfo\x12\x44\n\nsubstrates\x18\xc8\x01 \x03(\x0b\x32/.kpex.tech.CapacitanceInfo.SubstrateCapacitance\x12@\n\x08overlaps\x18\xc9\x01 \x03(\x0b\x32-.kpex.tech.CapacitanceInfo.OverlapCapacitance\x12\x42\n\tsidewalls\x18\xca\x01 \x03(\x0b\x32..kpex.tech.CapacitanceInfo.SidewallCapacitance\x12H\n\x0csideoverlaps\x18\xcb\x01 \x03(\x0b\x32\x31.kpex.tech.CapacitanceInfo.SideOverlapCapacitance\x1a\x63\n\x14SubstrateCapacitance\x12\x12\n\nlayer_name\x18\x01 \x01(\t\x12\x18\n\x10\x61rea_capacitance\x18\x02 \x01(\x01\x12\x1d\n\x15perimeter_capacitance\x18\x03 \x01(\x01\x1a\\\n\x12OverlapCapacitance\x12\x16\n\x0etop_layer_name\x18\x01 \x01(\t\x12\x19\n\x11\x62ottom_layer_name\x18\x02 \x01(\t\x12\x13\n\x0b\x63\x61pacitance\x18\x03 \x01(\x01\x1aN\n\x13SidewallCapacitance\x12\x12\n\nlayer_name\x18\x01 \x01(\t\x12\x13\n\x0b\x63\x61pacitance\x18\x02 \x01(\x01\x12\x0e\n\x06offset\x18\x03 \x01(\x01\x1a\\\n\x16SideOverlapCapacitance\x12\x15\n\rin_layer_name\x18\x01 \x01(\t\x12\x16\n\x0eout_layer_name\x18\x02 \x01(\t\x12\x13\n\x0b\x63\x61pacitance\x18\x03 \x01(\x01\"\x0e\n\x0cStyleVariantb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'process_parasitics_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_PROCESSPARASITICSINFO']._serialized_start=40
  _globals['_PROCESSPARASITICSINFO']._serialized_end=178
  _globals['_RESISTANCEINFO']._serialized_start=181
  _globals['_RESISTANCEINFO']._serialized_end=461
  _globals['_RESISTANCEINFO_LAYERRESISTANCE']._serialized_start=313
  _globals['_RESISTANCEINFO_LAYERRESISTANCE']._serialized_end=406
  _globals['_RESISTANCEINFO_VIARESISTANCE']._serialized_start=408
  _globals['_RESISTANCEINFO_VIARESISTANCE']._serialized_end=461
  _globals['_CAPACITANCEINFO']._serialized_start=464
  _globals['_CAPACITANCEINFO']._serialized_end=1128
  _globals['_CAPACITANCEINFO_SUBSTRATECAPACITANCE']._serialized_start=761
  _globals['_CAPACITANCEINFO_SUBSTRATECAPACITANCE']._serialized_end=860
  _globals['_CAPACITANCEINFO_OVERLAPCAPACITANCE']._serialized_start=862
  _globals['_CAPACITANCEINFO_OVERLAPCAPACITANCE']._serialized_end=954
  _globals['_CAPACITANCEINFO_SIDEWALLCAPACITANCE']._serialized_start=956
  _globals['_CAPACITANCEINFO_SIDEWALLCAPACITANCE']._serialized_end=1034
  _globals['_CAPACITANCEINFO_SIDEOVERLAPCAPACITANCE']._serialized_start=1036
  _globals['_CAPACITANCEINFO_SIDEOVERLAPCAPACITANCE']._serialized_end=1128
  _globals['_STYLEVARIANT']._serialized_start=1130
  _globals['_STYLEVARIANT']._serialized_end=1144
# @@protoc_insertion_point(module_scope)



#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX 
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#
from __future__ import annotations
from collections import defaultdict
from dataclasses import dataclass, field
from typing import *

import klayout_pex_protobuf.process_parasitics_pb2 as process_parasitics_pb2


NetName = str
LayerName = str
CellName = str


@dataclass
class NodeRegion:
    layer_name: LayerName
    net_name: NetName
    cap_to_gnd: float
    perimeter: float
    area: float


@dataclass(frozen=True)
class SidewallKey:
    layer: LayerName
    net1: NetName
    net2: NetName


@dataclass
class SidewallCap:  # see Magic EdgeCap, extractInt.c L444
    key: SidewallKey
    cap_value: float   # femto farad
    distance: float    # distance in µm
    length: float      # length in µm
    tech_spec: process_parasitics_pb2.CapacitanceInfo.SidewallCapacitance


@dataclass(frozen=True)
class OverlapKey:
    layer_top: LayerName
    net_top: NetName
    layer_bot: LayerName
    net_bot: NetName


@dataclass
class OverlapCap:
    key: OverlapKey
    cap_value: float  # femto farad
    shielded_area: float  # in µm^2
    unshielded_area: float  # in µm^2
    tech_spec: process_parasitics_pb2.CapacitanceInfo.OverlapCapacitance


@dataclass(frozen=True)
class SideOverlapKey:
    layer_inside: LayerName
    net_inside: NetName
    layer_outside: LayerName
    net_outside: NetName

    def __repr__(self) -> str:
        return f"{self.layer_inside}({self.net_inside})-"\
               f"{self.layer_outside}({self.net_outside})"


@dataclass
class SideOverlapCap:
    key: SideOverlapKey
    cap_value: float  # femto farad

    def __str__(self) -> str:
        return f"(Side Overlap): {self.key} = {round(self.cap_value, 6)}fF"


@dataclass(frozen=True)
class NetCoupleKey:
    net1: NetName
    net2: NetName

    def __repr__(self) -> str:
        return f"{self.net1}-{self.net2}"

    # NOTE: we norm net names alphabetically
    def normed(self) -> NetCoupleKey:
        if self.net1 < self.net2:
            return self
        else:
            return NetCoupleKey(self.net2, self.net1)


@dataclass
class ExtractionSummary:
    capacitances: Dict[NetCoupleKey, float]

    @classmethod
    def merged(cls, summaries: List[ExtractionSummary]) -> ExtractionSummary:
        merged_capacitances = defaultdict(float)
        for s in summaries:
            for couple_key, cap in s.capacitances.items():
                merged_capacitances[couple_key.normed()] += cap
        return ExtractionSummary(merged_capacitances)


@dataclass
class CellExtractionResults:
    cell_name: CellName

    overlap_coupling: Dict[OverlapKey, OverlapCap] = field(default_factory=dict)
    sidewall_table: Dict[SidewallKey, SidewallCap] = field(default_factory=dict)
    sideoverlap_table: Dict[SideOverlapKey, SideOverlapCap] = field(default_factory=dict)

    def summarize(self) -> ExtractionSummary:
        overlap_summary = ExtractionSummary({
            NetCoupleKey(key.net_top, key.net_bot): cap.cap_value
            for key, cap in self.overlap_coupling.items()
        })

        sidewall_summary = ExtractionSummary({
            NetCoupleKey(key.net1, key.net2): cap.cap_value
            for key, cap in self.sidewall_table.items()
        })

        sideoverlap_summary = ExtractionSummary({
            NetCoupleKey(key.net_inside, key.net_outside): cap.cap_value
            for key, cap in self.sideoverlap_table.items()
        })

        return ExtractionSummary.merged([
            overlap_summary, sidewall_summary, sideoverlap_summary
        ])


@dataclass
class ExtractionResults:
    cell_extraction_results: Dict[CellName, CellExtractionResults] = field(default_factory=dict)

    def summarize(self) -> ExtractionSummary:
        subsummaries = [s.summarize() for s in self.cell_extraction_results.values()]
        return ExtractionSummary.merged(subsummaries)


#! /usr/bin/env python3
#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX 
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#

import math
from collections import defaultdict
from typing import *

import klayout.db as kdb
import klayout.rdb as rdb

from ..klayout.lvsdb_extractor import KLayoutExtractionContext, GDSPair
from ..log import (
    console,
    debug,
    info,
    warning,
    error
)
from ..tech_info import TechInfo
from .extraction_results import *
import klayout_pex_protobuf.process_stack_pb2 as process_stack_pb2


EdgeInterval = Tuple[float, float]
ChildIndex = int
EdgeNeighborhood = List[Tuple[EdgeInterval, Dict[ChildIndex, List[kdb.Polygon]]]]


class RCExtractor:
    def __init__(self,
                 pex_context: KLayoutExtractionContext,
                 tech_info: TechInfo,
                 report_path: str):
        self.pex_context = pex_context
        self.tech_info = tech_info
        self.report_path = report_path

    def gds_pair(self, layer_name) -> Optional[GDSPair]:
        gds_pair = self.tech_info.gds_pair_for_computed_layer_name.get(layer_name, None)
        if not gds_pair:
            gds_pair = self.tech_info.gds_pair_for_layer_name.get(layer_name, None)
        if not gds_pair:
            warning(f"Can't find GDS pair for layer {layer_name}")
            return None
        return gds_pair

    def shapes_of_net(self, layer_name: str, net: kdb.Net) -> Optional[kdb.Region]:
        gds_pair = self.gds_pair(layer_name=layer_name)
        if not gds_pair:
            return None

        shapes = self.pex_context.shapes_of_net(gds_pair=gds_pair, net=net)
        if not shapes:
            debug(f"Nothing extracted for layer {layer_name}")
        return shapes

    def shapes_of_layer(self, layer_name: str) -> Optional[kdb.Region]:
        gds_pair = self.gds_pair(layer_name=layer_name)
        if not gds_pair:
            return None

        shapes = self.pex_context.shapes_of_layer(gds_pair=gds_pair)
        if not shapes:
            debug(f"Nothing extracted for layer {layer_name}")
        return shapes

    def extract(self) -> ExtractionResults:
        extraction_results = ExtractionResults()

        # TODO: for now, we always flatten and have only 1 cell
        cell_name = self.pex_context.top_cell.name
        report = rdb.ReportDatabase(f"PEX {cell_name}")
        cell_extraction_result = self.extract_cell(cell_name=cell_name, report=report)
        extraction_results.cell_extraction_results[cell_name] = cell_extraction_result

        report.save(self.report_path)

        return extraction_results

    def extract_cell(self,
                     cell_name: CellName,
                     report: rdb.ReportDatabase) -> CellExtractionResults:
        lvsdb = self.pex_context.lvsdb
        netlist: kdb.Netlist = lvsdb.netlist()
        dbu = self.pex_context.dbu

        extraction_results = CellExtractionResults(cell_name=cell_name)

        rdb_cell = report.create_cell(cell_name)
        rdb_cat_common = report.create_category("Common")
        rdb_cat_sidewall_old = report.create_category("Sidewall (legacy space_check)")
        rdb_cat_sidewall = report.create_category("Sidewall (EdgeNeighborhoodVisitor)")
        rdb_cat_overlap = report.create_category("Overlap")
        rdb_cat_fringe = report.create_category("Fringe / Side Overlap")

        def rdb_output(parent_category: rdb.RdbCategory,
                       category_name: str,
                       shapes: kdb.Shapes | kdb.Region | List[kdb.Edge]):
            rdb_cat = report.create_category(parent_category, category_name)
            report.create_items(rdb_cell.rdb_id(),  ## TODO: if later hierarchical mode is introduced
                                rdb_cat.rdb_id(),
                                kdb.CplxTrans(mag=dbu),
                                shapes)

        circuit = netlist.circuit_by_name(self.pex_context.top_cell.name)
        # https://www.klayout.de/doc-qt5/code/class_Circuit.html
        if not circuit:
            circuits = [c.name for c in netlist.each_circuit()]
            raise Exception(f"Expected circuit called {self.pex_context.top_cell.name} in extracted netlist, "
                            f"only available circuits are: {circuits}")

        #----------------------------------------------------------------------------------------
        layer2net2regions = defaultdict(lambda: defaultdict(kdb.Region))
        net2layer2regions = defaultdict(lambda: defaultdict(kdb.Region))
        layer_by_name: Dict[LayerName, process_stack_pb2.ProcessStackInfo.LayerInfo] = {}

        layer_regions_by_name: Dict[LayerName, kdb.Region] = defaultdict(kdb.Region)
        all_region = kdb.Region()
        regions_below_layer: Dict[LayerName, kdb.Region] = defaultdict(kdb.Region)
        regions_below_and_including_layer: Dict[LayerName, kdb.Region] = defaultdict(kdb.Region)
        all_layer_names: List[LayerName] = []
        layer_names_below: Dict[LayerName, List[LayerName]] = {}
        shielding_layer_names: Dict[Tuple[LayerName, LayerName], List[LayerName]] = defaultdict(list)
        previous_layer_name: Optional[str] = None

        substrate_region = kdb.Region()
        substrate_region.insert(self.pex_context.top_cell_bbox().enlarged(8.0 / dbu))  # 8 µm halo
        substrate_layer_name = self.tech_info.internal_substrate_layer_name
        layer_names_below[substrate_layer_name] = []
        all_layer_names.append(substrate_layer_name)
        layer2net2regions[substrate_layer_name][substrate_layer_name] = substrate_region
        net2layer2regions[substrate_layer_name][substrate_layer_name] = substrate_region
        layer_regions_by_name[substrate_layer_name] = substrate_region
        # NOTE: substrate not needed for
        #     - all_region
        #     - regions_below_layer
        #     - regions_below_and_including_layer

        for metal_layer in self.tech_info.process_metal_layers:
            layer_name = metal_layer.name
            gds_pair = self.gds_pair(layer_name)
            canonical_layer_name = self.tech_info.canonical_layer_name_by_gds_pair[gds_pair]

            all_layer_shapes = self.shapes_of_layer(layer_name) or kdb.Region()
            layer_regions_by_name[canonical_layer_name] += all_layer_shapes
            # NOTE: multiple LVS layers can be mapped to the same canonical name
            if previous_layer_name != canonical_layer_name:
                regions_below_layer[canonical_layer_name] += all_region
                layer_names_below[canonical_layer_name] = list(all_layer_names)
                for ln in all_layer_names:
                    lp = (canonical_layer_name, ln)
                    shielding_layer_names[lp] = [l for l in all_layer_names
                                                 if l != ln and l not in layer_names_below[ln]]
                    shielding_layer_names[ln, canonical_layer_name] = shielding_layer_names[lp]
                all_layer_names.append(canonical_layer_name)
            all_region += all_layer_shapes
            regions_below_and_including_layer[canonical_layer_name] += all_region

            previous_layer_name = canonical_layer_name

            for net in circuit.each_net():
                net_name = net.expanded_name()

                shapes = self.shapes_of_net(layer_name=layer_name, net=net)
                if shapes:
                    layer2net2regions[canonical_layer_name][net_name] += shapes
                    net2layer2regions[net_name][canonical_layer_name] += shapes
                    layer_by_name[canonical_layer_name] = metal_layer

        shielded_regions_between_layers: Dict[Tuple[LayerName, LayerName], kdb.Region] = {}
        for top_layer_name in layer2net2regions.keys():
            for bot_layer_name in reversed(layer_names_below[top_layer_name]):
                shielded_region = kdb.Region()
                shielding_layers = shielding_layer_names.get((top_layer_name, bot_layer_name), None)
                if shielding_layers:
                    for sl in shielding_layers:
                        shielded_region += layer_regions_by_name[sl]
                shielded_region.merge()
                shielded_regions_between_layers[(top_layer_name, bot_layer_name)] = shielded_region
                shielded_regions_between_layers[(bot_layer_name, top_layer_name)] = shielded_region
                if shielded_region:
                    rdb_output(rdb_cat_common, f"Shielded ({top_layer_name}-{bot_layer_name})", shielded_region)

        #----------------------------------------------------------------------------------------

        side_halo_um = self.tech_info.tech.process_parasitics.side_halo
        side_halo_dbu = int(side_halo_um / dbu) + 1  # add 1 nm to halo

        space_markers_by_layer_name: Dict[LayerName, kdb.Region] = {}
        rdb_cat_space_markers = report.create_category(rdb_cat_sidewall_old, "All Space Markers")

        for layer_name in layer2net2regions.keys():
            if layer_name == substrate_layer_name:
                continue

            space_markers = layer_regions_by_name[layer_name].space_check(
                d=side_halo_dbu,  # min space in um
                whole_edges=True,  # whole edges
                metrics=kdb.Metrics.Projection,  # metrics
                ignore_angle=None,  # ignore angle
                min_projection=None,  # min projection
                max_projection=None,  # max projection
                shielded=True,  # shielding
                opposite_filter=kdb.Region.NoOppositeFilter,  # error filter for opposite sides
                rect_filter=kdb.Region.NoRectFilter,  # error filter for rect input shapes
                negative=False,  # negative
                property_constraint=kdb.Region.IgnoreProperties,  # property_constraint
                zero_distance_mode=kdb.Region.IncludeZeroDistanceWhenTouching  # zero distance mode
            )
            space_markers_by_layer_name[layer_name] = space_markers
            rdb_output(rdb_cat_space_markers, f"layer={layer_name}", space_markers)

        #
        # (1) OVERLAP CAPACITANCE
        #
        for top_layer_name in layer2net2regions.keys():
            if top_layer_name == substrate_layer_name:
                continue

            top_net2regions = layer2net2regions.get(top_layer_name, None)
            if not top_net2regions:
                continue

            top_overlap_specs = self.tech_info.overlap_cap_by_layer_names.get(top_layer_name, None)
            if not top_overlap_specs:
                warning(f"No overlap cap specified for layer top={top_layer_name}")
                continue

            rdb_cat_top_layer = report.create_category(rdb_cat_overlap, f"top_layer={top_layer_name}")

            shapes_top_layer = layer_regions_by_name[top_layer_name]

            for bot_layer_name in reversed(layer_names_below[top_layer_name]):
                bot_net2regions = layer2net2regions.get(bot_layer_name, None)
                if not bot_net2regions:
                    continue

                overlap_cap_spec = top_overlap_specs.get(bot_layer_name, None)
                if not overlap_cap_spec:
                    warning(f"No overlap cap specified for layer top={top_layer_name}/bottom={bot_layer_name}")
                    continue

                rdb_cat_bot_layer = report.create_category(rdb_cat_top_layer, f"bot_layer={bot_layer_name}")

                shielded_region = shielded_regions_between_layers[(top_layer_name, bot_layer_name)].and_(shapes_top_layer)
                rdb_output(rdb_cat_bot_layer, "Shielded Between Layers Region", shielded_region)

                for net_top in top_net2regions.keys():
                    shapes_top_net: kdb.Region = top_net2regions[net_top].dup()

                    for net_bot in bot_net2regions.keys():
                        if net_top == net_bot:
                            continue

                        shapes_bot_net: kdb.Region = bot_net2regions[net_bot]

                        overlapping_shapes = shapes_top_net.__and__(shapes_bot_net)
                        if overlapping_shapes:
                            rdb_cat_nets = report.create_category(rdb_cat_bot_layer, f"{net_top} – {net_bot}")
                            rdb_output(rdb_cat_nets, "Overlapping Shapes", overlapping_shapes)

                            shielded_net_shapes = overlapping_shapes.__and__(shielded_region)
                            rdb_output(rdb_cat_nets, "Shielded Shapes", shielded_net_shapes)

                            unshielded_net_shapes = overlapping_shapes - shielded_net_shapes
                            rdb_output(rdb_cat_nets, "Unshielded Shapes", unshielded_net_shapes)

                            area_um2 = overlapping_shapes.area() * dbu**2
                            shielded_area_um2 = shielded_net_shapes.area() * dbu**2
                            unshielded_area_um2 = area_um2 - shielded_area_um2
                            cap_femto = unshielded_area_um2 * overlap_cap_spec.capacitance / 1000.0
                            shielded_cap_femto = shielded_area_um2 * overlap_cap_spec.capacitance / 1000.0
                            info(f"(Overlap): {top_layer_name}({net_top})-{bot_layer_name}({net_bot}): "
                                 f"Unshielded area: {unshielded_area_um2} µm^2, "
                                 f"cap: {round(cap_femto, 2)} fF")
                            if cap_femto > 0.0:
                                ovk = OverlapKey(layer_top=top_layer_name,
                                                 net_top=net_top,
                                                 layer_bot=bot_layer_name,
                                                 net_bot=net_bot)
                                cap = OverlapCap(key=ovk,
                                                 cap_value=cap_femto,
                                                 shielded_area=shielded_area_um2,
                                                 unshielded_area=unshielded_area_um2,
                                                 tech_spec=overlap_cap_spec)
                                report.create_category(  # used as info text
                                    rdb_cat_nets,
                                    f"{round(cap_femto, 3)} fF "
                                    f"({round(shielded_cap_femto, 3)} fF shielded "
                                    f"of total {round(cap_femto+shielded_cap_femto, 3)} fF)"
                                )
                                extraction_results.overlap_coupling[ovk] = cap

        # (2) SIDEWALL CAPACITANCE
        #
        for layer_name in layer2net2regions.keys():
            if layer_name == substrate_layer_name:
                continue

            sidewall_cap_spec = self.tech_info.sidewall_cap_by_layer_name.get(layer_name, None)
            if not sidewall_cap_spec:
                warning(f"No sidewall cap specified for layer {layer_name}")
                continue

            net2regions = layer2net2regions.get(layer_name, None)
            if not net2regions:
                continue

            rdb_cat_sw_layer = report.create_category(rdb_cat_sidewall_old, f"layer={layer_name}")

            space_markers = space_markers_by_layer_name[layer_name]

            for i, net1 in enumerate(net2regions.keys()):
                for j, net2 in enumerate(net2regions.keys()):
                    if i < j:

                        # info(f"Sidewall on {layer_name}: Nets {net1} <-> {net2}")
                        shapes1: kdb.Region = net2regions[net1]
                        shapes2: kdb.Region = net2regions[net2]

                        markers_net1 = space_markers.interacting(shapes1)
                        sidewall_edge_pairs = markers_net1.interacting(shapes2)

                        if not sidewall_edge_pairs:
                            continue

                        rdb_cat_sw_nets = report.create_category(rdb_cat_sidewall_old, f"{net1} - {net2}")
                        rdb_output(rdb_cat_sw_nets, f"Shapes {net1}", shapes1)
                        rdb_output(rdb_cat_sw_nets, f"Shapes {net2}", shapes2)
                        rdb_output(rdb_cat_sw_nets, f"Markers interacting {net1}", markers_net1)
                        rdb_output(rdb_cat_sw_nets, f"Markers interacting {net1}-{net2}", sidewall_edge_pairs)

                        for idx, pair in enumerate(sidewall_edge_pairs):
                            edge1: kdb.Edge = pair.first
                            edge2: kdb.Edge = pair.second

                            # TODO: support non-parallel situations
                            # avg_length = (edge1.length() + edge2.length()) / 2.0
                            # avg_distance = (pair.polygon(0).perimeter() - edge1.length() - edge2.length()) / 2.0
                            avg_length = min(edge1.length(), edge2.length())
                            avg_distance = pair.distance()

                            debug(f"Edge pair distance {avg_distance}, symmetric? {pair.symmetric}, "
                                 f"perimeter {pair.perimeter()}, parallel? {edge1.is_parallel(edge2)}")

                            # (3) SIDEWALL CAPACITANCE
                            #
                            # C = Csidewall * l * t / s
                            # C = Csidewall * l / s

                            length_um = avg_length * dbu
                            distance_um = avg_distance * dbu
                            
                            # NOTE: this is automatically bidirectional,
                            #       whereas MAGIC counts 2 sidewall contributions (one for each side of the cap)
                            cap_femto = (length_um * sidewall_cap_spec.capacitance) / \
                                        (distance_um + sidewall_cap_spec.offset) / 1000.0

                            rdb_output(rdb_cat_sw_nets, f"Edge Pair {idx}: {round(cap_femto, 3)} fF", pair)

                            info(f"(Sidewall) layer {layer_name}: Nets {net1} <-> {net2}: {round(cap_femto, 5)} fF")

                            swk = SidewallKey(layer=layer_name, net1=net1, net2=net2)
                            sw_cap = SidewallCap(key=swk,
                                                 cap_value=cap_femto,
                                                 distance=distance_um,
                                                 length=length_um,
                                                 tech_spec=sidewall_cap_spec)
                            extraction_results.sidewall_table[swk] = sw_cap

        #
        # (3) FRINGE / SIDE OVERLAP CAPACITANCE
        #

        class FringeEdgeNeighborhoodVisitor(kdb.EdgeNeighborhoodVisitor):
            def __init__(self,
                         inside_layer_name: str,
                         inside_net_name: str,
                         outside_layer_name: str,
                         child_names: List[str],
                         tech_info: TechInfo,
                         report_category: rdb.RdbCategory):
                self.inside_layer_name = inside_layer_name
                self.inside_net_name = inside_net_name
                self.outside_layer_name = outside_layer_name
                self.child_names = child_names
                # NOTE: child_names[0] is the inside net (foreign)
                #       child_names[1] is the shielded net (between layers)
                #       child_names[2:] are the outside nets
                self.tech_info = tech_info
                self.report_category = report_category

                # NOTE: overlap_cap_by_layer_names is top/bot (dict is not symmetric)
                self.overlap_cap_spec = tech_info.overlap_cap_by_layer_names[inside_layer_name].get(outside_layer_name, None)
                if not self.overlap_cap_spec:
                    self.overlap_cap_spec = tech_info.overlap_cap_by_layer_names[outside_layer_name][inside_layer_name]

                self.substrate_cap_spec = tech_info.substrate_cap_by_layer_name[inside_layer_name]
                self.sideoverlap_cap_spec = tech_info.side_overlap_cap_by_layer_names[inside_layer_name][outside_layer_name]

                self.sidewall_cap_spec = tech_info.sidewall_cap_by_layer_name[inside_layer_name]

                self.category_name_counter: Dict[str, int] = defaultdict(int)
                self.sidewall_counter = 0

            def begin_polygon(self,
                              layout: kdb.Layout,
                              cell: kdb.Cell,
                              polygon: kdb.Polygon):
                debug(f"----------------------------------------")
                debug(f"Polygon: {polygon}")

            def end_polygon(self):
                debug(f"End of polygon")

            def on_edge(self,
                        layout: kdb.Layout,
                        cell: kdb.Cell,
                        edge: kdb.Edge,
                        neighborhood: EdgeNeighborhood):
                #
                # NOTE: this complex operation will automatically rotate every edge to be on the x-axis
                #       going from 0 to edge.length
                #       so we only have to consider the y-axis to get the near and far distances
                #

                # TODO: consider z-shielding!

                debug(f"inside_layer={self.inside_layer_name}, "
                      f"inside_net={self.inside_net_name}, "
                      f"outside_layer={self.outside_layer_name}, "
                      f"edge = {edge}")

                rdb_inside_layer = report.create_category(rdb_cat_sidewall, f"layer={self.inside_layer_name}")
                rdb_sidewall_inside_net = report.create_category(rdb_inside_layer, f"inside={self.inside_net_name}")

                for (x1, x2), polygons_by_net in neighborhood:
                    if not polygons_by_net:
                        continue

                    edge_interval_length = x2 - x1
                    edge_interval_length_um = edge_interval_length * dbu

                    edge_interval_original = (self.to_original_trans(edge) *
                                              kdb.Edge(kdb.Point(x1, 0), kdb.Point(x2, 0)))
                    transformed_category_name = f"Edge interval {(x1, x2)}"
                    self.category_name_counter[transformed_category_name] += 1
                    rdb_cat_edge_interval = report.create_category(
                        self.report_category,
                        f"{transformed_category_name} ({self.category_name_counter[transformed_category_name]})"
                    )
                    rdb_output(rdb_cat_edge_interval, f"Original Section {edge_interval_original}", edge_interval_original)

                    polygons_on_same_layer = polygons_by_net.get(1, None)
                    shielded_region_lateral = kdb.Region()
                    if polygons_on_same_layer:
                        shielded_region_lateral.insert(polygons_on_same_layer)
                        rdb_output(rdb_cat_edge_interval, 'Laterally nearby shapes',
                                   kdb.Region([self.to_original_trans(edge) * p for p in shielded_region_lateral]))

                        # NOTE: first lateral nearby shape blocks everything beyond (like sidewall situation) up to halo
                        def distance_near(p: kdb.Polygon) -> float:
                            bbox: kdb.Box = p.bbox()

                            if not p.is_box():
                                warning(f"Side overlap, outside polygon {p} is not a box. "
                                        f"Currently, only boxes are supported, will be using bounding box {bbox}")
                            ## distance_near = (bbox.p1.y + bbox.p2.y) / 2.0
                            distance_near = min(bbox.p1.y, bbox.p2.y)
                            if distance_near < 0:
                                distance_near = 0
                            return distance_near

                        nearest_lateral_shape = (math.inf, polygons_on_same_layer[0])
                        for p in polygons_on_same_layer:
                            dnear = distance_near(p)
                            if dnear < nearest_lateral_shape[0]:
                                nearest_lateral_shape = (dnear, p)

                        rdb_output(rdb_cat_edge_interval, 'Closest nearby shape',
                                   kdb.Region(self.to_original_trans(edge) * nearest_lateral_shape[1]))

                        # NOTE: this method is always called for a single nearest edge (line), so the
                        #       polygons have 4 points.
                        #       Polygons points are sorted clockwise, so the edge
                        #       that goes from right-to-left is the nearest edge
                        nearby_opposing_edge = [e for e in nearest_lateral_shape[1].each_edge() if e.d().x < 0][-1]
                        nearby_opposing_edge_trans = self.to_original_trans(edge) * nearby_opposing_edge

                        opposing_net = '<unknown>'
                        # find the opposing net
                        for other_net, region in layer2net2regions[self.inside_layer_name].items():
                            if other_net == self.inside_net_name:
                                continue
                            if region.interacting(nearby_opposing_edge_trans).count() >= 1:
                                # we found the other net!
                                opposing_net = other_net
                                break

                        rdb_output(rdb_cat_edge_interval,
                                   f"Closest nearby edge (net {opposing_net})", [nearby_opposing_edge_trans])

                        sidewall_edge_pair = [nearby_opposing_edge_trans, edge_interval_original]
                        distance_um = nearest_lateral_shape[0] * dbu
                        sidewall_cap_femto = (edge_interval_length_um * self.sidewall_cap_spec.capacitance) / \
                                             (distance_um + self.sidewall_cap_spec.offset) / 1000.0 / 2.0

                        rdb_sidewall_outside_net = report.create_category(rdb_sidewall_inside_net,
                                                                         f"outside={opposing_net}")
                        self.sidewall_counter += 1
                        rdb_output(rdb_sidewall_outside_net,
                                   f"#{self.sidewall_counter}: "
                                   f"len {round(edge_interval_length_um, 3)} µm, distance {round(distance_um, 3)} µm, "
                                   f"{round(sidewall_cap_femto, 3)} fF",
                                   sidewall_edge_pair)

                        nearby_shield = kdb.Polygon([nearby_opposing_edge.p1,
                                                     nearby_opposing_edge.p2,
                                                     kdb.Point(x1, side_halo_dbu),
                                                     kdb.Point(x2, side_halo_dbu)])

                        rdb_output(rdb_cat_edge_interval, 'Nearby shield',
                                   kdb.Region(self.to_original_trans(edge) * nearby_shield))

                    shielded_region_between = kdb.Region()
                    shielded_polygons = polygons_by_net.get(2, None)  # shielded from layers between
                    if shielded_polygons:
                        shielded_region_between.insert(shielded_polygons)

                    for net_index, polygons in polygons_by_net.items():
                        if net_index == 0: # laterally shielded
                            continue
                        elif net_index == 1:  # ignore "shielded"
                            continue

                        if not polygons:
                            continue

                        unshielded_region: kdb.Region = kdb.Region(polygons) - shielded_region_between
                        if not unshielded_region:
                            continue

                        net_name = self.child_names[net_index]
                        rdb_cat_outside_net = report.create_category(rdb_cat_edge_interval,
                                                                     f"outside_net={net_name}")

                        rdb_output(rdb_cat_outside_net, 'Unshielded',
                                   kdb.Region([self.to_original_trans(edge) * p for p in unshielded_region]))

                        for p in unshielded_region:
                            bbox: kdb.Box = p.bbox()

                            if not p.is_box():
                                warning(f"Side overlap, outside polygon {p} is not a box. "
                                        f"Currently, only boxes are supported, will be using bounding box {bbox}")
                            distance_near = bbox.p1.y  #+ 1
                            if distance_near < 0:
                                distance_near = 0
                            distance_far = bbox.p2.y   #- 2
                            if distance_far < 0:
                                distance_far = 0
                            try:
                                assert distance_near >= 0
                                assert distance_far >= distance_near
                            except AssertionError:
                                print()
                                raise

                            if distance_far == distance_near:
                                continue

                            distance_near_um = distance_near * dbu
                            distance_far_um = distance_far * dbu

                            # NOTE: overlap scaling is 1/50  (see MAGIC ExtTech)
                            alpha_scale_factor = 0.02 * 0.01 * 0.5 * 200.0
                            alpha_c = self.overlap_cap_spec.capacitance * alpha_scale_factor

                            # see Magic ExtCouple.c L1164
                            cnear = (2.0 / math.pi) * math.atan(alpha_c * distance_near_um)
                            cfar = (2.0 / math.pi) * math.atan(alpha_c * distance_far_um)

                            # "cfrac" is the fractional portion of the fringe cap seen
                            # by tile tp along its length.  This is independent of the
                            # portion of the boundary length that tile tp occupies.
                            cfrac = cfar - cnear

                            # The fringe portion extracted from the substrate will be
                            # different than the portion added to the coupling layer.
                            sfrac: float

                            # see Magic ExtCouple.c L1198
                            alpha_s = self.substrate_cap_spec.area_capacitance / alpha_scale_factor
                            if alpha_s != alpha_c:
                                snear = (2.0 / math.pi) * math.atan(alpha_s * distance_near_um)
                                sfar = (2.0 / math.pi) * math.atan(alpha_s * distance_far_um)
                                sfrac = sfar - snear
                            else:
                                sfrac = cfrac

                            if outside_layer_name == substrate_layer_name:
                                cfrac = sfrac

                            cap_femto = (cfrac * edge_interval_length_um *
                                         self.sideoverlap_cap_spec.capacitance / 1000.0)
                            if cap_femto > 0.0:
                                report.create_category(rdb_cat_outside_net, f"{round(cap_femto, 3)} fF")  # used as info text

                                sok = SideOverlapKey(layer_inside=self.inside_layer_name,
                                                     net_inside=self.inside_net_name,
                                                     layer_outside=self.outside_layer_name,
                                                     net_outside=net_name)
                                sov = extraction_results.sideoverlap_table.get(sok, None)
                                if sov:
                                    sov.cap_value += cap_femto
                                else:
                                    sov = SideOverlapCap(key=sok, cap_value=cap_femto)
                                    extraction_results.sideoverlap_table[sok] = sov

                            # efflength = (cfrac - sov.so_coupfrac) * (double) length;
                            # cap += e->ec_cap * efflength;
                            #
                            # subfrac += sov.so_subfrac; / *Just add the shielded fraction * /
                            # efflength = (sfrac - subfrac) * (double) length;
                            #
                            # subcap = ExtCurStyle->exts_perimCap[ta][0] * efflength;

                            # TODO: shielding lateral

                            # TODO: fringe portion extracted from substrate

        for inside_layer_name in layer2net2regions.keys():
            if inside_layer_name == substrate_layer_name:
                continue

            inside_net2regions = layer2net2regions.get(inside_layer_name, None)
            if not inside_net2regions:
                continue

            inside_fringe_specs = self.tech_info.side_overlap_cap_by_layer_names.get(inside_layer_name, None)
            if not inside_fringe_specs:
                warning(f"No fringe / side overlap cap specified for layer inside={inside_layer_name}")
                continue

            shapes_inside_layer = layer_regions_by_name[inside_layer_name]
            fringe_halo_inside = shapes_inside_layer.sized(side_halo_dbu) - shapes_inside_layer

            rdb_cat_inside_layer = report.create_category(rdb_cat_fringe, f"inside_layer={inside_layer_name}")
            rdb_output(rdb_cat_inside_layer, "fringe_halo_inside", fringe_halo_inside)

            # Side Overlap: metal <-> metal (additionally, substrate)
            for outside_layer_name in layer2net2regions.keys():
                if inside_layer_name == outside_layer_name:
                    continue

                outside_net2regions = layer2net2regions.get(outside_layer_name, None)
                if not outside_net2regions:
                    continue

                cap_spec = inside_fringe_specs.get(outside_layer_name, None)
                if not cap_spec:
                    warning(f"No side overlap cap specified for "
                            f"layer inside={inside_layer_name}/outside={outside_layer_name}")
                    continue

                shapes_outside_layer = layer_regions_by_name[outside_layer_name]
                if not shapes_outside_layer:
                    continue

                shapes_outside_layer_within_halo = shapes_outside_layer.__and__(fringe_halo_inside)
                if not shapes_outside_layer_within_halo:
                    continue

                rdb_cat_outside_layer = report.create_category(rdb_cat_inside_layer,
                                                               f"outside_layer={outside_layer_name}")

                shielded_regions_between = shielded_regions_between_layers[(inside_layer_name, outside_layer_name)]
                rdb_output(rdb_cat_outside_layer, 'Shielded between layers', shielded_regions_between)

                for net_inside in inside_net2regions.keys():
                    shapes_inside_net: kdb.Region = inside_net2regions[net_inside]
                    if not shapes_inside_net:
                        continue

                    rdb_cat_inside_net = report.create_category(rdb_cat_outside_layer,
                                                                f"inside_net={net_inside}")

                    visitor = FringeEdgeNeighborhoodVisitor(
                        inside_layer_name=inside_layer_name,
                        inside_net_name=net_inside,
                        outside_layer_name=outside_layer_name,
                        child_names=[net_inside, 'NEARBY_SHAPES', 'SHIELD_BETWEEN'] +
                                    [k for k in outside_net2regions.keys() if k != net_inside],
                        tech_info=self.tech_info,
                        report_category=rdb_cat_inside_net
                    )

                    nearby_shapes = shapes_inside_layer - shapes_inside_net
                    # children = [kdb.CompoundRegionOperationNode.new_secondary(shapes_inside_net),
                    children = [kdb.CompoundRegionOperationNode.new_foreign(),
                                kdb.CompoundRegionOperationNode.new_secondary(nearby_shapes),
                                kdb.CompoundRegionOperationNode.new_secondary(shielded_regions_between)] + \
                               [kdb.CompoundRegionOperationNode.new_secondary(region)
                                for net, region in list(outside_net2regions.items())
                                if net != net_inside]

                    node = kdb.CompoundRegionOperationNode.new_edge_neighborhood(
                        children,
                        visitor,
                        0, # bext
                        0, # eext,
                        0, # din
                        side_halo_dbu # dout
                    )

                    shapes_inside_net.complex_op(node)

        for so in extraction_results.sideoverlap_table.values():
            info(so)

        return extraction_results



#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX 
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#
from __future__ import annotations

import tempfile
from typing import *
from dataclasses import dataclass
from rich.pretty import pprint

import klayout.db as kdb

import klayout_pex_protobuf.tech_pb2 as tech_pb2
from ..log import (
    console,
    debug,
    info,
    warning,
    error,
    rule
)

from ..tech_info import TechInfo


GDSPair = Tuple[int, int]


@dataclass
class KLayoutExtractedLayerInfo:
    index: int
    lvs_layer_name: str        # NOTE: this can be computed, so gds_pair is preferred
    gds_pair: GDSPair
    region: kdb.Region


@dataclass
class KLayoutMergedExtractedLayerInfo:
    source_layers: List[KLayoutExtractedLayerInfo]
    gds_pair: GDSPair


@dataclass
class KLayoutExtractionContext:
    lvsdb: kdb.LayoutToNetlist
    dbu: float
    top_cell: kdb.Cell
    layer_map: Dict[int, kdb.LayerInfo]
    cell_mapping: kdb.CellMapping
    target_layout: kdb.Layout
    extracted_layers: Dict[GDSPair, KLayoutMergedExtractedLayerInfo]
    unnamed_layers: List[KLayoutExtractedLayerInfo]

    @classmethod
    def prepare_extraction(cls,
                           lvsdb: kdb.LayoutToNetlist,
                           top_cell: str,
                           tech: TechInfo,
                           blackbox_devices: bool) -> KLayoutExtractionContext:
        dbu = lvsdb.internal_layout().dbu
        target_layout = kdb.Layout()
        target_layout.dbu = dbu
        top_cell = target_layout.create_cell(top_cell)

        # CellMapping
        #   mapping of internal layout to target layout for the circuit mapping
        #   https://www.klayout.de/doc-qt5/code/class_CellMapping.html
        # ---
        # https://www.klayout.de/doc-qt5/code/class_LayoutToNetlist.html#method18
        # Creates a cell mapping for copying shapes from the internal layout to the given target layout
        cm = lvsdb.cell_mapping_into(target_layout,  # target layout
                                     top_cell,
                                     not blackbox_devices)  # with_device_cells

        lm = cls.build_LVS_layer_map(target_layout=target_layout,
                                     lvsdb=lvsdb,
                                     tech=tech,
                                     blackbox_devices=blackbox_devices)

        net_name_prop_num = 1

        # Build a full hierarchical representation of the nets
        # https://www.klayout.de/doc-qt5/code/class_LayoutToNetlist.html#method14
        # hier_mode = None
        hier_mode = kdb.LayoutToNetlist.BuildNetHierarchyMode.BNH_Flatten
        # hier_mode = kdb.LayoutToNetlist.BuildNetHierarchyMode.BNH_SubcircuitCells

        lvsdb.build_all_nets(
            cmap=cm,               # mapping of internal layout to target layout for the circuit mapping
            target=target_layout,  # target layout
            lmap=lm,               # maps: target layer index => net regions
            hier_mode=hier_mode,   # hier mode
            netname_prop=net_name_prop_num,  # property name to which to attach the net name
            circuit_cell_name_prefix="CIRCUIT_",
            device_cell_name_prefix=None  # "DEVICE_"
        )

        extracted_layers, unnamed_layers = cls.nonempty_extracted_layers(lvsdb=lvsdb,
                                                                         tech=tech,
                                                                         blackbox_devices=blackbox_devices)

        return KLayoutExtractionContext(
            lvsdb=lvsdb,
            dbu=dbu,
            top_cell=top_cell,
            layer_map=lm,
            cell_mapping=cm,
            target_layout=target_layout,
            extracted_layers=extracted_layers,
            unnamed_layers=unnamed_layers
        )

    @staticmethod
    def build_LVS_layer_map(target_layout: kdb.Layout,
                            lvsdb: kdb.LayoutToNetlist,
                            tech: TechInfo,
                            blackbox_devices: bool) -> Dict[int, kdb.LayerInfo]:
        # NOTE: currently, the layer numbers are auto-assigned
        # by the sequence they occur in the LVS script, hence not well defined!
        # build a layer map for the layers that correspond to original ones.

        # https://www.klayout.de/doc-qt5/code/class_LayerInfo.html
        lm: Dict[int, kdb.LayerInfo] = {}

        if not hasattr(lvsdb, "layer_indexes"):
            raise Exception("Needs at least KLayout version 0.29.2")

        for layer_index in lvsdb.layer_indexes():
            lname = lvsdb.layer_name(layer_index)

            computed_layer_info = tech.computed_layer_info_by_name.get(lname, None)
            if computed_layer_info and blackbox_devices:
                match computed_layer_info.kind:
                    case tech_pb2.ComputedLayerInfo.Kind.KIND_DEVICE_RESISTOR:
                        continue
                    case tech_pb2.ComputedLayerInfo.Kind.KIND_DEVICE_CAPACITOR:
                        continue

            gds_pair = tech.gds_pair_for_computed_layer_name.get(lname, None)
            if not gds_pair:
                li = lvsdb.internal_layout().get_info(layer_index)
                if li != kdb.LayerInfo():
                    gds_pair = (li.layer, li.datatype)

            if gds_pair is not None:
                target_layer_index = target_layout.layer(*gds_pair)  # Creates a new internal layer!
                region = lvsdb.layer_by_index(layer_index)
                lm[target_layer_index] = region

        return lm

    @staticmethod
    def nonempty_extracted_layers(lvsdb: kdb.LayoutToNetlist,
                                  tech: TechInfo,
                                  blackbox_devices: bool) -> Tuple[Dict[GDSPair, KLayoutMergedExtractedLayerInfo], List[KLayoutExtractedLayerInfo]]:
        # https://www.klayout.de/doc-qt5/code/class_LayoutToNetlist.html#method18
        nonempty_layers: Dict[GDSPair, KLayoutMergedExtractedLayerInfo] = {}

        unnamed_layers: List[KLayoutExtractedLayerInfo] = []

        for idx, ln in enumerate(lvsdb.layer_names()):
            layer = lvsdb.layer_by_name(ln)
            if layer.count() >= 1:
                computed_layer_info = tech.computed_layer_info_by_name.get(ln, None)
                if not computed_layer_info:
                    warning(f"Unable to find info about extracted LVS layer '{ln}'")
                    gds_pair = (1000 + idx, 20)
                    linfo = KLayoutExtractedLayerInfo(
                        index=idx,
                        lvs_layer_name=ln,
                        gds_pair=gds_pair,
                        region=layer
                    )
                    unnamed_layers.append(linfo)
                    continue

                if blackbox_devices:
                    match computed_layer_info.kind:
                        case tech_pb2.ComputedLayerInfo.Kind.KIND_DEVICE_RESISTOR:
                            continue
                        case tech_pb2.ComputedLayerInfo.Kind.KIND_DEVICE_CAPACITOR:
                            continue

                gds_pair = (computed_layer_info.layer_info.gds_layer, computed_layer_info.layer_info.gds_datatype)

                linfo = KLayoutExtractedLayerInfo(
                    index=idx,
                    lvs_layer_name=ln,
                    gds_pair=gds_pair,
                    region=layer
                )

                entry = nonempty_layers.get(gds_pair, None)
                if entry:
                    entry.source_layers.append(linfo)
                else:
                    nonempty_layers[gds_pair] = KLayoutMergedExtractedLayerInfo(
                        source_layers=[linfo],
                        gds_pair=gds_pair,
                    )

        return nonempty_layers, unnamed_layers

    def top_cell_bbox(self) -> kdb.Box:
        b1: kdb.Box = self.target_layout.top_cell().bbox()
        b2: kdb.Box = self.lvsdb.internal_layout().top_cell().bbox()
        if b1.area() > b2.area():
            return b1
        else:
            return b2

    def shapes_of_net(self, gds_pair: GDSPair, net: kdb.Net) -> Optional[kdb.Region]:
        lyr = self.extracted_layers.get(gds_pair, None)
        if not lyr:
            return None

        shapes: kdb.Region

        match len(lyr.source_layers):
            case 0:
                raise AssertionError('Internal error: Empty list of source_layers')
            case 1:
                shapes = self.lvsdb.shapes_of_net(net, lyr.source_layers[0].region, True)
            case _:
                shapes = kdb.Region()
                for sl in lyr.source_layers:
                    shapes += self.lvsdb.shapes_of_net(net, sl.region, True)
                # shapes.merge()

        return shapes

    def shapes_of_layer(self, gds_pair: GDSPair) -> Optional[kdb.Region]:
        lyr = self.extracted_layers.get(gds_pair, None)
        if not lyr:
            return None

        shapes: kdb.Region

        match len(lyr.source_layers):
            case 0:
                raise AssertionError('Internal error: Empty list of source_layers')
            case 1:
                shapes = lyr.source_layers[0].region
            case _:
                shapes = kdb.Region()
                for sl in lyr.source_layers:
                    shapes += sl.region
                # shapes.merge()

        return shapes





#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX 
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#
from __future__ import annotations
from enum import IntEnum
from functools import cached_property
import logging
import rich.console
import rich.logging
from typing import *


class LogLevel(IntEnum):
    ALL = 0
    DEBUG = 10
    SUBPROCESS = 12
    VERBOSE = 15
    INFO = 20
    WARNING = 30
    ERROR = 40
    CRITICAL = 50
    DEFAULT = SUBPROCESS

    @classmethod
    @cached_property
    def level_by_name(cls) -> Dict[str, LogLevel]:
        return {e.name: e for e in cls}


class LogLevelFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        msg = record.getMessage()
        match record.levelno:
            case LogLevel.WARNING.value: return f"[yellow]{msg}"
            case LogLevel.ERROR.value: return f"[red]{msg}"
            case _:
                return msg


class LogLevelFilter(logging.Filter):
    def __init__(self, levels: Iterable[str], invert: bool = False):
        super().__init__()
        self.levels = levels
        self.invert = invert

    def filter(self, record: logging.LogRecord) -> bool:
        if self.invert:
            return record.levelname not in self.levels
        else:
            return record.levelname in self.levels


console = rich.console.Console()
__logger = logging.getLogger("__kpex__")


def set_log_level(log_level: LogLevel):
    __logger.setLevel(log_level)


def register_additional_handler(handler: logging.Handler):
    """
    Adds a new handler to the default logger.

    :param handler: The new handler. Must be of type ``logging.Handler``
        or its subclasses.
    """
    __logger.addHandler(handler)


def deregister_additional_handler(handler: logging.Handler):
    """
    Removes a registered handler from the default logger.

    :param handler: The handler. If not registered, the behavior
        of this function is undefined.
    """
    __logger.removeHandler(handler)



def configure_logger():
    global __logger, console

    for level in LogLevel:
        logging.addLevelName(level=level.value, levelName=level.name)

    subprocess_rich_handler = rich.logging.RichHandler(
        console=console,
        show_time=False,
        omit_repeated_times=False,
        show_level=False,
        show_path=False,
        enable_link_path=False,
        markup=False,
        tracebacks_word_wrap=False,
        keywords=[]
    )
    subprocess_rich_handler.addFilter(LogLevelFilter(['SUBPROCESS']))

    rich_handler = rich.logging.RichHandler(
        console=console,
        omit_repeated_times=False,
        show_level=True,
        markup=True,
        rich_tracebacks=True,
        tracebacks_suppress=[],
        keywords=[]
    )

    rich_handler.setFormatter(LogLevelFormatter(fmt='%(message)s', datefmt='[%X]'))
    rich_handler.addFilter(LogLevelFilter(['SUBPROCESS'], invert=True))

    set_log_level(LogLevel.SUBPROCESS)

    __logger.handlers.clear()
    __logger.addHandler(subprocess_rich_handler)
    __logger.addHandler(rich_handler)


def debug(*args, **kwargs):
    if not kwargs.get('stacklevel'):  # ensure logged file location is correct
        kwargs['stacklevel'] = 2
    __logger.debug(*args, **kwargs)


def subproc(msg: object, **kwargs):
    if not kwargs.get('stacklevel'):  # ensure logged file location is correct
        kwargs['stacklevel'] = 2
    __logger.log(LogLevel.SUBPROCESS, msg, **kwargs)


def rule(title: str = '', **kwargs):  # pragma: no cover
    """
    Prints a horizontal line on the terminal enclosing the first argument
    if the log level is <= INFO.

    Kwargs are passed to https://rich.readthedocs.io/en/stable/reference/console.html#rich.console.Console.rule

    :param title: A title string to enclose in the console rule
    """
    console.rule(title)


def info(*args, **kwargs):
    if not kwargs.get('stacklevel'):  # ensure logged file location is correct
        kwargs['stacklevel'] = 2
    __logger.info(*args, **kwargs)


def warning(*args, **kwargs):
    if not kwargs.get('stacklevel'):  # ensure logged file location is correct
        kwargs['stacklevel'] = 2
    __logger.warning(*args, **kwargs)


def error(*args, **kwargs):
    if not kwargs.get('stacklevel'):  # ensure logged file location is correct
        kwargs['stacklevel'] = 2
    __logger.error(*args, **kwargs)


configure_logger()


#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX 
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#
from __future__ import annotations

import copy
from dataclasses import dataclass
from typing import *


@dataclass
class CapacitanceMatrix:
    conductor_names: List[str]  # NOTE FasterCap generates [g_1, g_2, ...]
    rows: List[List[float]]     # NOTE: in µm

    def __getitem__(self, key):
        return self.rows.__getitem__(key)

    def __setitem__(self, key, value):
        self.rows.__setitem__(key, value)

    @property
    def dimension(self):
        return len(self.conductor_names)

    @classmethod
    def parse_csv(cls, path: str, separator: str = ';'):
        with open(path, 'r') as f:
            lines = f.readlines()
            if len(lines) < 2:
                raise Exception(f"Capacitance Matrix CSV must at least have 2 lines: "
                                f"{path}")
            conductor_names = [cell.strip() for cell in lines[0].split(sep=separator)]
            rows = []
            for line in lines[1:]:
                row = [float(cell.strip()) for cell in line.split(sep=separator)]
                rows.append(row)
            return CapacitanceMatrix(conductor_names=conductor_names,
                                     rows=rows)

    def write_csv(self, output_path: str, separator: str = ';'):
        with open(output_path, 'w') as f:
            header_line = separator.join(self.conductor_names)
            f.write(header_line)
            f.write('\n')

            for row in self.rows:
                cells = ['%.12g' % cell for cell in row]
                row_line = separator.join(cells)
                f.write(row_line)
                f.write('\n')

    def averaged_off_diagonals(self) -> CapacitanceMatrix:
        c = copy.deepcopy(self)
        for i in range(len(self.rows)):
            for j in range(len(self.conductor_names)):
                if j <= i:
                    continue
                v1 = self[i][j]
                v2 = self[j][i]
                avg = (v1 + v2) / 2
                # print(f"i={i} j={j}, avg({v1}, {v2}) == {avg}")
                c[i][j] = avg
                c[j][i] = avg
        return c



#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX 
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#
from __future__ import annotations

import os
import subprocess
import time

from ..log import (
    debug,
    info,
    warning,
    error,
    subproc,
    rule
)


class LVSRunner:
    @staticmethod
    def run_klayout_lvs(exe_path: str,
                        lvs_script: str,
                        gds_path: str,
                        schematic_path: str,
                        log_path: str,
                        lvsdb_path: str,
                        verbose: bool):
        args = [
            exe_path,
            '-b',
            '-r', lvs_script,
            '-rd', f"input={os.path.abspath(gds_path)}",
            '-rd', f"report={os.path.abspath(lvsdb_path)}",
            '-rd', f"schematic={os.path.abspath(schematic_path)}",
            '-rd', 'thr=22',
            '-rd', 'run_mode=deep',
            '-rd', 'spice_net_names=true',
            '-rd', 'spice_comments=false',
            '-rd', 'scale=false',
            '-rd', f"verbose={'true' if verbose else 'false'}",
            '-rd', 'schematic_simplify=false',
            '-rd', 'net_only=false',
            '-rd', 'top_lvl_pins=true',
            '-rd', 'combine=false',
            '-rd', 'combine_devices=false', # IHP
            '-rd', 'purge=false',
            '-rd', 'purge_nets=false',
            '-rd', 'no_simplify=true', # IHP
        ]
        rule('Calling KLayout LVS script')
        subproc(' '.join(args))
        subproc(log_path)
        start = time.time()

        proc = subprocess.Popen(args,
                                stdin=subprocess.DEVNULL,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT,
                                universal_newlines=True,
                                text=True)
        with open(log_path, 'w') as f:
            while True:
                line = proc.stdout.readline()
                if not line:
                    break
                subproc(line[:-1])  # remove newline
                f.writelines([line])
        proc.wait()

        duration = time.time() - start

        rule()

        if proc.returncode == 0:
            info(f"klayout LVS succeeded after {'%.4g' % duration}s")
        else:
            warning(f"klayout LVS failed with status code {proc.returncode} after {'%.4g' % duration}s, "
                    f"see log file: {log_path}")



#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX 
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#
import re
import time
from typing import *

import os
import subprocess

from ..log import (
    debug,
    info,
    warning,
    error,
    rule,
    subproc,
)
from ..common.capacitance_matrix import CapacitanceMatrix


def run_fastcap(exe_path: str,
                lst_file_path: str,
                log_path: str,
                expansion_order: float = 2,
                partitioning_depth: str = 'auto',
                permittivity_factor: float = 1.0,
                iterative_tolerance: float = 0.01):
    work_dir = os.path.dirname(lst_file_path)

    # we have to chdir into the directory containing the lst file,
    # so make all paths absolute, and the lst_file relative
    log_path = os.path.abspath(log_path)
    lst_file_path = os.path.basename(lst_file_path)

    info(f"Chdir to {work_dir}")
    os.chdir(work_dir)
    args = [
        exe_path,
        f"-o{expansion_order}",
    ]

    if partitioning_depth != 'auto':
        args += [
            f"-d{partitioning_depth}",
        ]

    args += [
        f"-p{permittivity_factor}",
        f"-t{iterative_tolerance}",
        f"-l{lst_file_path}",
    ]

    info(f"Calling FastCap2")
    subproc(f"{' '.join(args)}, output file: {log_path}")

    rule('FastCap Output')
    start = time.time()

    proc = subprocess.Popen(args,
                            stdin=subprocess.DEVNULL,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT,
                            universal_newlines=True,
                            text=True)
    with open(log_path, 'w') as f:
        while True:
            line = proc.stdout.readline()
            if not line:
                break
            subproc(line[:-1])  # remove newline
            f.writelines([line])
    proc.wait()

    duration = time.time() - start

    rule()

    if proc.returncode == 0:
        info(f"FastCap2 succeeded after {'%.4g' % duration}s")
    else:
        raise Exception(f"FastCap2 failed with status code {proc.returncode} after {'%.4g' % duration}s, "
                        f"see log file: {log_path}")


# CAPACITANCE MATRIX, picofarads
#                     1          2          3          4
# $1%GROUP2 1       7850      -7277      -2115      54.97
# $1%GROUP2 2      -7277  3.778e+05      130.9 -3.682e+05
# $2%GROUP3 3      -2115      130.9       6792      -5388
# $2%GROUP3 4      54.97 -3.682e+05      -5388  3.753e+05
def fastcap_parse_capacitance_matrix(log_path: str) -> CapacitanceMatrix:
    with open(log_path, 'r') as f:
        rlines = f.readlines()
        rlines.reverse()

        # multiple iterations possible, find the last matrix
        for idx, line in enumerate(rlines):
            if line.startswith('CAPACITANCE MATRIX, '):
                section_m = re.match(r'CAPACITANCE MATRIX, (\w+)', line)
                if not section_m:
                    raise Exception(f"Could not parse capacitor unit")
                unit_str = section_m.group(1)

                dimension_line = rlines[idx-1].strip()
                dimensions = dimension_line.split()  # remove whitespace
                dim = len(dimensions)
                conductor_names: List[str] = []
                rows: List[List[float]] = []
                for i in reversed(range(idx-1-dim, idx-1)):
                    line = rlines[i].strip()
                    cells = [cell.strip() for cell in line.split(' ')]
                    if cells[1] != str(i):
                        warning(f"Expected capacitor matrix row to have index {i}, but obtained {cells[1]}")
                    cells.pop(1)
                    cells = list(filter(lambda c: len(c) >= 1, cells))
                    conductor_names.append(cells[0])
                    row = [float(cell)/1e6 for cell in cells[1:]]
                    rows.append(row)
                cm = CapacitanceMatrix(conductor_names=conductor_names, rows=rows)
                return cm

        raise Exception(f"Could not extract capacitance matrix from FasterCap log file {log_path}")



#! /usr/bin/env python3
#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#

from typing import *
from pathlib import Path
import re

from .magic_ext_data_structures import *


def parse_magic_ext_file(path: Path) -> ExtData:
    ports: List[Port] = []
    nodes: List[Node] = []
    devices: List[Device] = []

    with open(path, 'r') as f:
        for line in f.readlines():
            # port "VDD" 2 -600 800 500 1000 m1
            # port "VSS" 3 -600 500 500 700 m1
            m = re.match(
                r'^port "(?P<net>\w+)" (?P<nr>\d+) (?P<x_bot>-?\d+) (?P<y_bot>-?\d+) (?P<x_top>-?\d+) (?P<y_top>-?\d+) (?P<layer>\w+)$',
                line.strip())
            if m:
                ports.append(Port(net=m.group('net'),
                                  x_bot=int(m.group('x_bot')),
                                  y_bot=int(m.group('y_bot')),
                                  x_top=int(m.group('x_top')),
                                  y_top=int(m.group('y_top')),
                                  layer=m.group('layer')))

            m = re.match(
                r'^(node|substrate) "(?P<net>\w+)" (?P<int_r>\d+) (?P<fin_c>\d+) (?P<x_bot>-?\d+) (?P<y_bot>-?\d+) (?P<layer>\w+) .*$',
                line.strip())
            if m:
                nodes.append(Node(net=m.group('net'),
                                  int_r=int(m.group('int_r')),
                                  fin_c=int(m.group('fin_c')),
                                  x_bot=int(m.group('x_bot')),
                                  y_bot=int(m.group('y_bot')),
                                  layer=m.group('layer')))

            m = re.match(
                r'^device (?P<type>\w+) (?P<model>\w+) (?P<x_bot>-?\d+) (?P<y_bot>-?\d+) (?P<x_top>-?\d+) (?P<y_top>-?\d+) .*$',
                line.strip())
            if m:
                t = m.group('type')
                device_type = DeviceType(t)
                devices.append(Device(device_type=device_type,
                                      model=m.group('model'),
                                      x_bot=int(m.group('x_bot')),
                                      y_bot=int(m.group('y_bot')),
                                      x_top=int(m.group('x_top')),
                                      y_top=int(m.group('y_top'))))
    return ExtData(ports=ports,
                   nodes=nodes,
                   devices=devices)


def parse_magic_res_ext_file(path: Path) -> ResExtData:
    rnodes = []
    resistors = []

    with open(path, 'r') as f:
        for line in f.readlines():
            m = re.match(
                r'^rnode "(?P<name>[\w\\.]+)" (?P<int_r>-?\d+) (?P<fin_c>-?\d+) (?P<x_bot>-?\d+) (?P<y_bot>-?\d+) 0$',
                line.strip())
            if m:
                rnodes.append(ResNode(name=m.group('name'),
                                      int_r=int(m.group('int_r')),  # NOTE: R is always 0!
                                      fin_c=int(m.group('fin_c')),
                                      x_bot=int(m.group('x_bot')),
                                      y_bot=int(m.group('y_bot'))))

            m = re.match(
                r'^resist "(?P<node1>[\w\\.]+)" "(?P<node2>[\w\\.]+)" (?P<value>-?[\d\\.]+)$',
                line.strip())
            if m:
                resistors.append(Resistor(node1=m.group('node1'),
                                          node2=m.group('node2'),
                                          value_ohm=float(m.group('value'))))

    return ResExtData(rnodes=rnodes, resistors=resistors)



#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX 
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#

# A class providing a service for building FastCap2 or FasterCap models
#
# This class is used the following way:
#
# 1) Create a FasterCapModelBuilder object
#    Specify the default k value which is the k assumed for "empty space".
#    You can also specify a maximum area and the "b" parameter for the
#    triangulation. The b parameter corresponds to the minimum angle
#    and should be <=1 (b=sin(min_angle)*2).
#    I.e. b=1 -> min_angle=30 deg, b=0.5 -> min_angle~14.5 deg.
#
# 2) Add material definitions for the dielectrics
#    Each material definition consists of a k value and
#    a material name.
#
# 3) Add layers in the 2.5d view fashion
#    Each layer is a sheet in 3d space that is extruded in vertical
#    direction with the given start and stop z (or height)
#    The layer must be a DRC::Layer or RBA::Region object.
#
#    Layers can be added in two ways:
#
#    * As conductors: specify the net name
#
#    * As dielectric layer: specify the material name
#
#    The layers can intersect. The package resolves intersections
#    based on priority: conductors first, dielectrics according to
#    their position in the "materials" definition (first entries have
#    higher prio)
#
# 4) Generate a 3d model using "generate"
#    This method returns an object you can use to generate STL files
#    or FastCap files.


from __future__ import annotations

import base64
from collections import defaultdict
import hashlib
import os
from typing import *
from dataclasses import dataclass
from functools import reduce
import math

import klayout.db as kdb

from ..log import (
    debug,
    info,
    warning,
    error,
    subproc
)


@dataclass
class FasterCapModelBuilder:
    dbu: float
    """Database unit"""

    k_void: float
    """Default dielectric of 'empty space'"""

    delaunay_amax: float
    """Maximum area parameter for the Delaunay triangulation"""

    delaunay_b: float
    """
    The delaunay_b parameter for the Delaunay triangulation 
    corresponds to the minimum angle
    and should be <=1 (b=sin(min_angle)*2).
    I.e. b=1 -> min_angle=30 deg, b=0.5 -> min_angle~14.5 deg.
    """

    def __init__(self,
                 dbu: float,
                 k_void: float,
                 delaunay_amax: float = 0.0,
                 delaunay_b: float = 1.0,
                 ):
        self.dbu = dbu
        self.k_void = k_void
        self.delaunay_amax = delaunay_amax
        self.delaunay_b = delaunay_b

        self.materials: Dict[str, float] = {}
        self.net_names: List[str] = []

        #                           layer,            zstart, zstop
        self.clayers: Dict[str, List[Tuple[kdb.Region, float, float]]] = {}
        self.dlayers: Dict[str, List[Tuple[kdb.Region, float, float]]] = {}

        info(f"DBU: {'%.12g' % self.dbu}")
        info(f"Delaunay b: {'%.12g' % self.delaunay_b}")
        info(f"Delaunay area_max: {'%.12g' % self.delaunay_amax}")

    def add_material(self, name: str, k: float):
        self.materials[name] = k

    def add_dielectric(self,
                       material_name: str,
                       layer: kdb.Region,
                       z: float,
                       height: float):
        if hasattr(layer, 'data'):
            layer = layer.data
        self._add_layer(name=material_name, layer=layer, is_dielectric=True, z=z, height=height)

    def add_conductor(self,
                      net_name: str,
                      layer: kdb.Region,
                      z: float,
                      height: float):
        if hasattr(layer, 'data'):
            layer = layer.data
        self._add_layer(name=net_name, layer=layer, is_dielectric=False, z=z, height=height)

    def _norm2z(self, z: float) -> float:
        return z * self.dbu

    def _z2norm(self, z: float) -> float:
        return math.floor(z / self.dbu + 1e-6)

    def _add_layer(self,
                   name: str,
                   layer: kdb.Region,
                   z: float,
                   height: float,
                   is_dielectric: bool):
        if is_dielectric and name not in self.materials:
            raise ValueError(f"Unknown material {name} - did you use 'add_material'?")

        zstart: float = z
        zstop: float = zstart + height

        if is_dielectric:
            if name not in self.dlayers:
                self.dlayers[name] = []
            self.dlayers[name].append((layer, self._z2norm(zstart), self._z2norm(zstop)))
        else:
            if name not in self.clayers:
                self.clayers[name] = []
            self.clayers[name].append((layer, self._z2norm(zstart), self._z2norm(zstop)))

    def generate(self) -> Optional[FasterCapModelGenerator]:
        z: List[float] = []
        for ll in (self.dlayers, self.clayers):
            for k, v in ll.items():
                for l in v:
                    z.extend((l[1], l[2]))
        z = sorted([*{*z}])  # sort & uniq
        if len(z) == 0:
            return None

        gen = FasterCapModelGenerator(dbu=self.dbu,
                                      k_void=self.k_void,
                                      delaunay_amax=self.delaunay_amax,
                                      delaunay_b=self.delaunay_b,
                                      materials=self.materials,
                                      net_names=list(self.clayers.keys()))
        for zcurr in z:
            gen.next_z(self._norm2z(zcurr))

            for nn, v in self.clayers.items():
                for l in v:
                    if l[1] <= zcurr < l[2]:
                        gen.add_in(name=f"+{nn}", layer=l[0])
                    if l[1] < zcurr <= l[2]:
                        gen.add_out(name=f"+{nn}", layer=l[0])
            for mn, v in self.dlayers.items():
                for l in v:
                    if l[1] <= zcurr < l[2]:
                        gen.add_in(name=f"-{mn}", layer=l[0])
                    if l[1] < zcurr <= l[2]:
                        gen.add_out(name=f"-{mn}", layer=l[0])

            gen.finish_z()

        gen.finalize()
        return gen


@dataclass(frozen=True)
class HDielKey:
    outside: Optional[str]
    inside: Optional[str]

    def __str__(self) -> str:
        return f"{self.outside or 'void'} <-> {self.inside or 'void'}"

    @property
    def topic(self) -> str:
        return 'dielectric'

    def reversed(self) -> HDielKey:
        return HDielKey(self.inside, self.outside)


@dataclass(frozen=True)
class HCondKey:
    net_name: str
    outside: Optional[str]

    def __str__(self) -> str:
        return f"{self.outside or 'void'} <-> {self.net_name}"

    @property
    def topic(self) -> str:
        return 'conductor'


@dataclass(frozen=True)
class VKey:
    kk: HDielKey | HCondKey
    p0: kdb.DPoint
    de: kdb.DVector


@dataclass(frozen=True)
class Point:
    x: float
    y: float
    z: float

    def __sub__(self, other: Point) -> Point:
        return Point(self.x - other.x, self.y - other.y, self.z - other.z)

    def sq_length(self) -> float:
        return self.x**2 + self.y**2 + self.z**2

    def to_fastcap(self) -> str:
        return '%.12g %.12g %.12g' % (self.x, self.y, self.z)


def vector_product(a: Point, b: Point) -> Point:
    vp = Point(
        a.y * b.z - a.z * b.y,
        a.z * b.x - a.x * b.z,
        a.x * b.y - a.y * b.x
    )
    return vp


def dot_product(a: Point, b: Point) -> float:
    dp = a.x * b.x + a.y * b.y + a.z * b.z
    return dp


@dataclass(frozen=True)
class Triangle:
    p0: Point
    p1: Point
    p2: Point

    def reversed(self) -> Triangle:
        return Triangle(self.p2, self.p1, self.p0)

    def outside_reference_point(self) -> Point:
        v1 = self.p1 - self.p0
        v2 = self.p2 - self.p0
        vp = Point(v1.y  * v2.z - v1.z * v2.y,
                   -v1.x * v2.z + v1.z * v2.x,
                   v1.x  * v2.y - v1.y * v2.x)
        vp_abs = math.sqrt(vp.x ** 2 + vp.y ** 2 + vp.z ** 2)
        rp = Point(self.p0.x + vp.x / vp_abs,
                   self.p0.y + vp.y / vp_abs,
                   self.p0.z + vp.z / vp_abs)
        return rp

    def to_fastcap(self) -> str:
        return ' '.join([p.to_fastcap() for p in (self.p0, self.p1, self.p2)])

    def __len__(self):
        return 3

    def __getitem__(self, i) -> Point:
        match i:
            case 0: return self.p0
            case 1: return self.p1
            case 2: return self.p2
            case _: raise IndexError("list index out of range")


@dataclass(frozen=True)
class Edge:
    p0: Point
    p1: Point

    def vector_of_edge(self) -> Point:
        return Point(
            self.p1.x - self.p0.x,
            self.p1.y - self.p0.y,
            self.p1.z - self.p0.z
        )

    def reversed(self) -> Edge:
        return Edge(self.p1, self.p0)


@dataclass
class FasterCapModelGenerator:
    dbu: float
    """Database unit"""

    k_void: float
    """Default dielectric of 'empty space'"""

    delaunay_amax: float
    """Maximum area parameter for the Delaunay triangulation"""

    delaunay_b: float
    """
    The delaunay_b parameter for the Delaunay triangulation 
    corresponds to the minimum angle
    and should be <=1 (b=sin(min_angle)*2).
    I.e. b=1 -> min_angle=30 deg, b=0.5 -> min_angle~14.5 deg.
    """

    materials: Dict[str, float]
    """Maps material name to dielectric k"""

    net_names: List[str]

    def __init__(self,
                 dbu: float,
                 k_void: float,
                 delaunay_amax: float,
                 delaunay_b: float,
                 materials: Dict[str, float],
                 net_names: List[str]):
        self.k_void = k_void
        self.delaunay_amax = delaunay_amax
        self.delaunay_b = delaunay_b
        self.dbu = dbu
        self.materials = materials
        self.net_names = net_names

        self.z: Optional[float] = None
        self.zz: Optional[float] = None
        self.layers_in: Dict[str, kdb.Region] = {}
        self.layers_out: Dict[str, kdb.Region] = {}
        self.state: Dict[str, kdb.Region] = {}
        self.current: Dict[str, List[kdb.Region]] = {}
        self.diel_data: Dict[HDielKey, List[Triangle]] = {}
        self.diel_vdata: Dict[VKey, kdb.Region] = {}
        self.cond_data: Dict[HCondKey, List[Triangle]] = {}
        self.cond_vdata: Dict[VKey, kdb.Region] = {}

    def reset(self):
        self.layers_in = {}
        self.layers_out = {}

    def add_in(self, name: str, layer: kdb.Region):
        debug(f"add_in: {name} -> {layer}")
        if name not in self.layers_in:
            # NOTE: Calling kdb.Region([]) with the empty list enforces, that a flat layer is created.
            #       The following "+=" packs the polynomials into this flat layer and
            #       does not copy the hierarchical layer from the LayoutToNetlist database.
            self.layers_in[name] = kdb.Region([])
        self.layers_in[name] += layer

    def add_out(self, name: str, layer: kdb.Region):
        debug(f"add_out: {name} -> {layer}")
        if name not in self.layers_out:
            # NOTE: Calling kdb.Region([]) with the empty list enforces, that a flat layer is created.
            #       The following "+=" packs the polynomials into this flat layer and
            #       does not copy the hierarchical layer from the LayoutToNetlist database.
            self.layers_out[name] = kdb.Region([])
        self.layers_out[name] += layer

    def finish_z(self):
        debug(f"Finishing layer z={self.z}")

        din: Dict[str, kdb.Region] = {}
        dout: Dict[str, kdb.Region] = {}
        all_in = kdb.Region()
        all_out = kdb.Region()
        all = kdb.Region()
        all_cin: Optional[kdb.Region] = None
        all_cout: Optional[kdb.Region] = None

        for names, prefix in ((self.net_names, '+'), (self.materials.keys(), '-')):
            for nn in names:
                mk = prefix + nn

                # compute merged events
                if mk not in self.current:
                    self.current[mk] = []
                current_before = self.current[mk][0].dup() if len(self.current[mk]) >= 1 else kdb.Region()
                lin, lout, current = self._merge_events(pyra=self.current[mk],
                                                        lin=self.layers_in.get(mk, None),
                                                        lout=self.layers_out.get(mk, None))
                debug(f"Merged events & status for {mk}:")
                debug(f"  in = {lin}")
                debug(f"  out = {lout}")
                debug(f"  state = {current}")

                if mk not in self.state:
                    self.state[mk] = kdb.Region()

                # legalize in and out events
                lin_org = lin.dup()
                lout_org = lout.dup()
                lout &= self.state[mk]
                lin -= all
                lout += current & all_in
                lin += current_before & all_out
                lin -= lout_org
                lout -= lin_org

                # tracks the legalized horizontal cuts
                self.state[mk] += lin
                self.state[mk] -= lout

                din[mk] = lin
                dout[mk] = lout

                debug(f"Legalized events & status for '{mk}':")
                debug(f"  in = {din[mk]}")
                debug(f"  out = {dout[mk]}")
                debug(f"  state = {self.state[mk]}")

                all_in += lin
                all_out += lout
                all += self.state[mk]

            if prefix == '+':
                all_cin = all_in.dup()
                all_cout = all_out.dup()

        debug(f"All conductor region in: {all_cin}")
        debug(f"All conductor region out: {all_cout}")

        # check whether states are separated
        a = reduce(lambda x, y: x+y, self.state.values())
        for k, s in self.state.items():
            r: kdb.Region = s - a
            if not r.is_empty():
                error(f"State region of {k} ({s}) is not contained entirely "
                      f"in remaining all state region ({a}) - this means there is an overlap")
            a -= s

        # Now we have legalized the in and out events
        for mni in self.materials.keys():
            lin = din.get(f"-{mni}", None)
            if lin:
                lin = lin.dup()
                lin -= all_cout  # handled with the conductor
                for mno in self.materials.keys():
                    lout = dout.get(f"-{mno}", None)
                    if lout:
                        d: kdb.Region = lout & lin
                        if not d.is_empty():
                            self.generate_hdiel(below=mno, above=mni, layer=d)
                        lin -= lout
                if not lin.is_empty():
                    self.generate_hdiel(below=None, above=mni, layer=lin)

        for mno in self.materials.keys():
            lout = dout.get(f"-{mno}", None)
            if lout:
                lout = lout.dup()
                lout -= all_cin  # handled with the conductor
                for mni in self.materials.keys():
                    lin = din.get(f"-{mni}", None)
                    if lin:
                        lout -= lin
                if not lout.is_empty():
                    self.generate_hdiel(below=mno, above=None, layer=lout)

        for nn in self.net_names:
            lin = din.get(f"+{nn}", None)
            if lin:
                lin = lin.dup()
                for mno in self.materials.keys():
                    lout = dout.get(f"-{mno}", None)
                    if lout:
                        d = lout & lin
                        if not d.is_empty():
                            self.generate_hcond_in(net_name=nn, below=mno, layer=d)
                        lin -= lout
                if not lin.is_empty():
                    self.generate_hcond_in(net_name=nn, below=None, layer=lin)

        for nn in self.net_names:
            lout = dout.get(f"+{nn}", None)
            if lout:
                lout = lout.dup()
                lout -= all_cin  # handled with the conductor
                for mni in self.materials.keys():
                    lin = din.get(f"-{mni}", None)
                    if lin:
                        d = lout & lin
                        if not d.is_empty():
                            self.generate_hcond_out(net_name=nn, above=mni, layer=d)
                        lout -= lin
                if not lout.is_empty():
                    self.generate_hcond_out(net_name=nn, above=None, layer=lout)

    def next_z(self, z: float):
        debug(f"Next layer {z}")

        self.reset()

        if self.z is None:
            self.z = z
            return

        self.zz = z

        all_cond = kdb.Region()
        for nn in self.net_names:
            mk = f"+{nn}"
            if mk in self.state:
                all_cond += self.state[mk]
        all_cond = all_cond.edges()

        for i, mni in enumerate(self.materials):
            linside = self.state.get(f"-{mni}", None)
            if linside:
                linside = linside.edges()
                linside -= all_cond  # handled with the conductor
                for o, mno in enumerate(self.materials):
                    if i != o:
                        loutside = self.state.get(f"-{mno}", None)
                        if loutside:
                            loutside = loutside.edges()
                            if o > i:
                                d = loutside & linside
                                for e in d:
                                    # NOTE: we need to swap points as we started from "outside"
                                    self.generate_vdiel(outside=mno, inside=mni, edge=e.swapped_points())
                            linside -= loutside

                for e in linside:
                    self.generate_vdiel(outside=None, inside=mni, edge=e)

        for nn in self.net_names:
            mk = f"+{nn}"
            linside = self.state.get(mk, None)
            if linside:
                linside = linside.edges()
                for mno in self.materials:
                    loutside = self.state.get(f"-{mno}", None)
                    if loutside:
                        loutside = loutside.edges()
                        d = loutside & linside
                        for e in d:
                            # NOTE: we need to swap points as we started from "outside"
                            self.generate_vcond(net_name=nn, outside=mno, edge=e.swapped_points())
                        linside -= loutside
                for e in linside:
                    self.generate_vcond(net_name=nn, outside=None, edge=e)

        self.z = z

    def generate_hdiel(self,
                       below: Optional[str],
                       above: Optional[str],
                       layer: kdb.Region):
        k = HDielKey(below, above)
        debug(f"Generating horizontal dielectric surface {k} as {layer}")
        if k not in self.diel_data:
            self.diel_data[k] = []
        data = self.diel_data[k]

        for t in layer.delaunay(self.delaunay_amax / self.dbu ** 2, self.delaunay_b):
            # NOTE: normal is facing downwards (to "below")
            pl = list(map(lambda pt: Point(pt.x * self.dbu, pt.y * self.dbu, self.z),
                           t.each_point_hull()))
            tri = Triangle(*pl)
            data.append(tri)
            debug(f"  {tri}")

    def generate_v_surface(self,
                           kk: HDielKey | HCondKey,
                           edge: kdb.Edge) -> Tuple[VKey, kdb.Box]:
        debug(f"Generating vertical {kk.topic} surface {kk}  with edge {edge}")

        el = math.sqrt(edge.sq_length())
        de = kdb.DVector(edge.d().x / el, edge.d().y / el)
        ne = kdb.DVector(edge.d().y / el, -edge.d().x / el)
        p0 = ne * ne.sprod(kdb.DPoint(edge.p1) - kdb.DPoint()) + kdb.DPoint()
        x1 = (edge.p1 - p0).sprod(de)
        x2 = (edge.p2 - p0).sprod(de)

        key = VKey(kk, p0, de)
        surface = kdb.Box(x1,
                          math.floor(self.z / self.dbu + 0.5),
                          x2,
                          math.floor(self.zz / self.dbu + 0.5))
        return key, surface

    def generate_vdiel(self,
                       outside: Optional[str],
                       inside: Optional[str],
                       edge: kdb.Edge):
        if edge.is_degenerate():
            return

        key, surface = self.generate_v_surface(HDielKey(outside, inside), edge)
        if key not in self.diel_vdata:
            self.diel_vdata[key] = kdb.Region()

        self.diel_vdata[key].insert(surface)

    def generate_hcond_in(self,
                          net_name: str,
                          below: Optional[str],
                          layer: kdb.Region):
        k = HCondKey(net_name, below)
        debug(f"Generating horizontal bottom conductor surface {k} as {layer}")

        if k not in self.cond_data:
            self.cond_data[k] = []
        data = self.cond_data[k]

        for t in layer.delaunay(self.delaunay_amax / self.dbu ** 2, self.delaunay_b):
            # NOTE: normal is facing downwards (to "below")
            pl = list(map(lambda pt: Point(pt.x * self.dbu, pt.y * self.dbu, self.z),
                           t.each_point_hull()))
            tri = Triangle(*pl)
            data.append(tri)
            debug(f"  {tri}")

    def generate_hcond_out(self,
                           net_name: str,
                           above: Optional[str],
                           layer: kdb.Region):
        k = HCondKey(net_name, above)
        debug(f"Generating horizontal top conductor surface {k} as {layer}")

        if k not in self.cond_data:
            self.cond_data[k] = []
        data = self.cond_data[k]

        for t in layer.delaunay(self.delaunay_amax / self.dbu ** 2, self.delaunay_b):
            # NOTE: normal is facing downwards (into conductor)
            pl = list(map(lambda pt: Point(pt.x * self.dbu, pt.y * self.dbu, self.z),
                           t.each_point_hull()))
            tri = Triangle(*pl)
            # now it is facing outside (to "above")
            tri = tri.reversed()
            data.append(tri)
            debug(f"  {tri}")

    def generate_vcond(self,
                       net_name: str,
                       outside: Optional[str],
                       edge: kdb.Edge):
        if edge.is_degenerate():
            return

        key, surface = self.generate_v_surface(HCondKey(net_name, outside), edge)
        if key not in self.cond_vdata:
            self.cond_vdata[key] = kdb.Region()

        self.cond_vdata[key].insert(surface)

    def triangulate(self, p0: kdb.DPoint, de: kdb.DVector, region: kdb.Region, data: List[Triangle]):
        def convert_point(pt: kdb.Point) -> Point:
            pxy = (p0 + de * pt.x) * self.dbu
            pz = pt.y * self.dbu
            return Point(pxy.x, pxy.y, pz)

        for t in region.delaunay(self.delaunay_amax / self.dbu ** 2, self.delaunay_b):
            # NOTE: normal is facing outwards (to "left")
            pl = list(map(convert_point, t.each_point_hull()))
            tri = Triangle(*pl)
            # now it is facing outside (to "above")
            data.append(tri)
            debug(f"  {tri}")

    def finalize(self):
        for k, r in self.diel_vdata.items():
            debug(f"Finishing vertical dielectric plane {k.kk} at {k.p0}/{k.de}")

            if k.kk not in self.diel_data:
                self.diel_data[k.kk] = []
            data = self.diel_data[k.kk]

            self.triangulate(p0=k.p0, de=k.de, region=r, data=data)

        for k, r in self.cond_vdata.items():
            debug(f"Finishing vertical conductor plane {k.kk} at {k.p0} / {k.de}")

            if k.kk not in self.cond_data:
                self.cond_data[k.kk] = []
            data = self.cond_data[k.kk]

            self.triangulate(p0=k.p0, de=k.de, region=r, data=data)

        dk: Dict[HDielKey, List[Triangle]] = {}

        for k in self.diel_data.keys():
            kk = k.reversed()
            if kk not in dk:
                dk[k] = []
            else:
                debug(f"Combining dielectric surfaces {kk} with reverse")

        for k, v in self.diel_data.items():
            kk = k.reversed()
            if kk in dk:
                dk[kk] += list(map(lambda t: t.reversed(), v))
            else:
                dk[k] += v

        self.diel_data = dk

    def write_fastcap(self, output_dir_path: str, prefix: str) -> str:
        max_filename_length: Optional[int] = None
        try:
            max_filename_length = os.pathconf(output_dir_path, 'PC_NAME_MAX')
        except AttributeError:
            pass  # NOTE: windows does not support the os.pathconf attribute

        lst_fn = os.path.join(output_dir_path, f"{prefix}.lst")
        file_num = 0
        lst_file: List[str] = [f"* k_void={'%.12g' % self.k_void}"]

        for k, data in self.diel_data.items():
            if len(data) == 0:
                continue

            file_num += 1

            k_outside = self.materials[k.outside] if k.outside else self.k_void
            k_inside = self.materials[k.inside] if k.inside else self.k_void

            # lst_file.append(f"* Dielectric interface: outside={outside}, inside={inside}")

            fn = f"{prefix}{file_num}_outside={k.outside or '(void)'}_inside={k.inside or '(void)'}.geo"
            output_path = os.path.join(output_dir_path, fn)
            self._write_fastercap_geo(output_path=output_path,
                                      data=data,
                                      cond_name=None,
                                      cond_number=file_num,
                                      rename_conductor=False)

            # NOTE: for now, we compute the reference points for each triangle
            #       This is a FasterCap feature, reference point in the *.geo file (end of each T line)
            rp_s = "0 0 0"
            lst_file.append(f"D {fn} {'%.12g' % k_outside} {'%.12g' % k_inside} 0 0 0 {rp_s}")

        #
        # Feedback from FastFieldSolvers:
        #
        # - using the '+' trailing statements (conductor collation),
        #   only the same conductor should be collated
        #
        # - renaming different conductor numbers ('N' rule line) is not allowed (currently a bug)
        #   - Example: 1->VDD (1.geo) and 2->VDD (2.geo) is not possible
        #   - Both conductor *.geo files should have the same number
        #   - only the last conductor *.geo file should contain the 'N' rule
        #
        # - reference points
        #
        cond_data_grouped_by_net = defaultdict(list)
        for k, data in self.cond_data.items():
            if len(data) == 0:
                continue
            cond_data_grouped_by_net[k.net_name].append((k.outside, data))

        cond_num = file_num

        for nn, cond_list in cond_data_grouped_by_net.items():
            cond_num += 1
            last_cond_index = len(cond_list) - 1
            for idx, (outside, data) in enumerate(cond_list):
                file_num += 1
                k_outside = self.materials[outside] if outside else self.k_void

                outside = outside or '(void)'
                # lst_file.append(f"* Conductor interface: outside={outside}, net={nn}")
                fn = f"{prefix}{file_num}_outside={outside}_net={nn}.geo"
                if max_filename_length is not None and len(fn) > max_filename_length:
                    warning(f"Unusual long net name detected: {nn}")
                    d = hashlib.md5(nn.encode('utf-8')).digest()
                    h = base64.urlsafe_b64encode(d).decode('utf-8').rstrip('=')
                    remaining_len = len(f"{prefix}_{file_num}_outside={outside}_net=.geo")
                    short_nn = nn[0: (max_filename_length - remaining_len - len(h) - 1)] + f"_{h}"
                    fn = f"{prefix}{file_num}_outside={outside}_net={short_nn}.geo"
                output_path = os.path.join(output_dir_path, fn)
                self._write_fastercap_geo(output_path=output_path,
                                          data=data,
                                          cond_number=cond_num,
                                          cond_name=nn,
                                          rename_conductor=(idx == last_cond_index))
                collation_operator = '' if idx == last_cond_index else ' +'
                lst_file.append(f"C {fn}  {'%.12g' % k_outside}  0 0 0{collation_operator}")

        subproc(lst_fn)
        with open(lst_fn, "w") as f:
            f.write('\n'.join(lst_file))
            f.write('\n')

        return lst_fn

    @staticmethod
    def _write_fastercap_geo(output_path: str,
                             data: List[Triangle],
                             cond_number: int,
                             cond_name: Optional[str],
                             rename_conductor: bool):
        subproc(output_path)
        with open(output_path, "w") as f:
            f.write(f"0 GEO File\n")
            for t in data:
                f.write(f"T {cond_number}")
                f.write(' ' + t.to_fastcap())

                # compute a reference point "outside"
                rp = t.outside_reference_point()
                rp_s = rp.to_fastcap()

                f.write(f" {rp_s}\n")
            if cond_name and rename_conductor:
                f.write(f"N {cond_number} {cond_name}\n")

    def check(self):
        info("Checking …")
        errors = 0

        for mn in self.materials.keys():
            tris = self._collect_diel_tris(mn)
            info(f"Material {mn} -> {len(tris)} triangles")
            errors += self._check_tris(f"Material '{mn}'", tris)

        for nn in self.net_names:
            tris = self._collect_cond_tris(nn)
            info(f"Net '{nn}' -> {len(tris)} triangles")
            errors += self._check_tris(f"Net '{nn}'", tris)

        if errors == 0:
            info("  No errors found")
        else:
            info(f"  {errors} error{'s' if errors >= 2 else ''} found")

    def _check_tris(self, msg: str, triangles: List[Triangle]) -> int:
        errors = 0

        edge_set: Set[Edge] = set()
        edges = self._normed_edges(triangles)

        for e in edges:
            if e in edge_set:
                error(f"{msg}: duplicate edge {self._edge2s(e)}")
                errors += 1
            else:
                edge_set.add(e)

        self._split_edges(edge_set)

        for e in edge_set:
            if e.reversed() not in edge_set:
                error(f"{msg}: edge {self._edge2s(e)} not connected with reverse edge (open surface)")
                errors += 1

        return errors

    def _normed_edges(self, triangles: List[Triangle]) -> List[Edge]:
        edges = []

        def normed_dbu(p: Point):
            return Point(*tuple(map(lambda c: math.floor(c / self.dbu + 0.5),
                                    (p.x, p.y, p.z))))

        for t in triangles:
            for i in range(0, 3):
                p1 = normed_dbu(t[i])
                p2 = normed_dbu(t[(i + 1) % 3])
                edges.append(Edge(p1, p2))

        return edges

    def _point2s(self, p: Point) -> str:
        return f"(%.12g, %.12g, %.12g)" % (p.x * self.dbu, p.y * self.dbu, p.z * self.dbu)

    def _edge2s(self, e: Edge) -> str:
        return f"{self._point2s(e.p0)}-{self._point2s(e.p1)}"

    @staticmethod
    def _is_antiparallel(a: Point,
                         b: Point) -> bool:
        vp = vector_product(a, b)
        if abs(vp.sq_length()) > 0.5:  # we got normalized!
            return False

        sp = dot_product(a, b)
        return sp < 0

    def _split_edges(self, edges: set[Edge]):
        edges_by_p2: DefaultDict[Point, List[Edge]] = defaultdict(list)
        edges_by_p1: DefaultDict[Point, List[Edge]] = defaultdict(list)
        for e in edges:
            edges_by_p2[e.p1].append(e)
            edges_by_p1[e.p0].append(e)

        i = 0
        while True:
            i += 1
            subst: DefaultDict[Edge, List[Edge]] = defaultdict(list)

            for e in edges:
                ee = edges_by_p2.get(e.p0, [])
                for eee in ee:
                    ve = e.vector_of_edge()
                    veee = eee.vector_of_edge()
                    if self._is_antiparallel(ve, veee) and \
                       (veee.sq_length() < ve.sq_length() - 0.5):
                        # There is a shorter edge antiparallel ->
                        # this means we need to insert a split point into e
                        subst[e] += [Edge(e.p0, eee.p0), Edge(eee.p0, e.p1)]

            for e in edges:
                ee = edges_by_p1.get(e.p1, [])
                for eee in ee:
                    ve = e.vector_of_edge()
                    veee = eee.vector_of_edge()
                    if self._is_antiparallel(ve, veee) and \
                       (veee.sq_length() < ve.sq_length() - 0.5):
                        # There is a shorter edge antiparallel ->
                        # this means we need to insert a split point into e
                        subst[e] += [Edge(e.p0, eee.p1), Edge(eee.p1, e.p1)]

            if len(subst) == 0:
                break

            for e, replacement in subst.items():
                edges_by_p1[e.p0].remove(e)
                edges_by_p2[e.p1].remove(e)
                edges.remove(e)
                for r in replacement:
                    edges.add(r)
                    edges_by_p1[r.p0].append(r)
                    edges_by_p2[r.p1].append(r)

    def dump_stl(self, output_dir_path: str, prefix: str):
        for mn in self.materials.keys():
            tris = self._collect_diel_tris(mn)
            output_path = os.path.join(output_dir_path, f"{prefix}diel_{mn}.stl")
            self._write_as_stl(output_path, tris)

        for nn in self.net_names:
            tris = self._collect_cond_tris(nn)
            output_path = os.path.join(output_dir_path, f"{prefix}cond_{nn}.stl")
            self._write_as_stl(output_path, tris)

    @staticmethod
    def _write_as_stl(file_name: str,
                      tris: List[Triangle]):
        if len(tris) == 0:
            return

        subproc(file_name)
        with open(file_name, "w") as f:
            f.write("solid stl\n")
            for t in tris:
                f.write("  facet normal 0 0 0\n")
                f.write("    outer loop\n")
                t = t.reversed()
                for p in (t.p0, t.p1, t.p2):
                    f.write(f"   vertex {p.to_fastcap()}\n")
                f.write("  endloop\n")
                f.write(" endfacet\n")
            f.write("endsolid stl\n")

    @staticmethod
    def _merge_events(pyra: List[Optional[kdb.Region]],
                      lin: Optional[kdb.Region],
                      lout: Optional[kdb.Region]) -> Tuple[kdb.Region, kdb.Region, kdb.Region]:
        lin = lin.dup() if lin else kdb.Region()
        lout = lout.dup() if lout else kdb.Region()
        past = pyra[0].dup() if len(pyra) >= 1 else kdb.Region()

        for i in range(0, len(pyra)):
            ii = len(pyra) - i
            added: kdb.Region = lin & pyra[ii - 1]
            if not added.is_empty():
                if ii >= len(pyra):
                    pyra.append(kdb.Region())
                    assert len(pyra) == ii + 1
                pyra[ii] += added
                lin -= added

        if len(pyra) == 0:
            pyra.append(kdb.Region())
        pyra[0] += lin

        for i in range(0, len(pyra)):
            ii = len(pyra) - i
            removed: kdb.Region = lout & pyra[ii - 1]
            if not removed.is_empty():
                pyra[ii - 1] -= removed
                lout -= removed

        # compute merged events
        lin = pyra[0] - past
        lout = past - pyra[0]
        return lin, lout, pyra[0]

    def _collect_diel_tris(self, material_name: str) -> List[Triangle]:
        tris = []

        for k, v in self.diel_data.items():
            if material_name == k.outside:
                tris += v
            elif material_name == k.inside:
                tris += [t.reversed() for t in v]

        for k, v in self.cond_data.items():
            if material_name == k.outside:
                tris += v

        return tris

    def _collect_cond_tris(self, net_name: str) -> List[Triangle]:
        tris = []
        for k, v in self.cond_data.items():
            if k.net_name == net_name:
                tris += [t.reversed() for t in v]
        return tris



#! /usr/bin/env python3
#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX 
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#


#
# Protocol Buffer Schema for FasterCap Input Files
# https://www.fastfieldsolvers.com/software.htm#fastercap
#

from typing import *
from functools import cached_property
import math

import klayout.db as kdb

from ..klayout.lvsdb_extractor import KLayoutExtractionContext, GDSPair
from .fastercap_model_generator import FasterCapModelBuilder, FasterCapModelGenerator
from ..log import (
    console,
    debug,
    info,
    warning,
    error
)
from ..tech_info import TechInfo

import klayout_pex_protobuf.process_stack_pb2 as process_stack_pb2


class FasterCapInputBuilder:
    def __init__(self,
                 pex_context: KLayoutExtractionContext,
                 tech_info: TechInfo,
                 k_void: float = 3.5,
                 delaunay_amax: float = 0.0,
                 delaunay_b: float = 1.0):
        self.pex_context = pex_context
        self.tech_info = tech_info
        self.k_void = k_void
        self.delaunay_amax = delaunay_amax
        self.delaunay_b = delaunay_b

    @cached_property
    def dbu(self) -> float:
        return self.pex_context.dbu

    def gds_pair(self, layer_name) -> Optional[GDSPair]:
        gds_pair = self.tech_info.gds_pair_for_computed_layer_name.get(layer_name, None)
        if not gds_pair:
            gds_pair = self.tech_info.gds_pair_for_layer_name.get(layer_name, None)
        if not gds_pair:
            warning(f"Can't find GDS pair for layer {layer_name}")
            return None
        return gds_pair

    def shapes_of_net(self, layer_name: str, net: kdb.Net) -> Optional[kdb.Region]:
        gds_pair = self.gds_pair(layer_name=layer_name)
        if not gds_pair:
            return None

        shapes = self.pex_context.shapes_of_net(gds_pair=gds_pair, net=net)
        if not shapes:
            debug(f"Nothing extracted for layer {layer_name}")
        return shapes

    def shapes_of_layer(self, layer_name: str) -> Optional[kdb.Region]:
        gds_pair = self.gds_pair(layer_name=layer_name)
        if not gds_pair:
            return None

        shapes = self.pex_context.shapes_of_layer(gds_pair=gds_pair)
        if not shapes:
            debug(f"Nothing extracted for layer {layer_name}")
        return shapes

    def top_cell_bbox(self) -> kdb.Box:
        return self.pex_context.top_cell_bbox()

    def build(self) -> FasterCapModelGenerator:
        lvsdb = self.pex_context.lvsdb
        netlist: kdb.Netlist = lvsdb.netlist()

        def format_terminal(t: kdb.NetTerminalRef) -> str:
            td = t.terminal_def()
            d = t.device()
            return f"{d.expanded_name()}/{td.name}/{td.description}"

        model_builder = FasterCapModelBuilder(
            dbu=self.dbu,
            k_void=self.k_void,
            delaunay_amax=self.delaunay_amax,   # test/compare with smaller, e.g. 0.05 => more triangles
            delaunay_b=self.delaunay_b          # test/compare with 1.0 => more triangles at edges
        )

        fox_layer = self.tech_info.field_oxide_layer

        model_builder.add_material(name=fox_layer.name, k=fox_layer.field_oxide_layer.dielectric_k)
        for diel_name, diel_k in self.tech_info.dielectric_by_name.items():
            model_builder.add_material(name=diel_name, k=diel_k)

        circuit = netlist.circuit_by_name(self.pex_context.top_cell.name)
        # https://www.klayout.de/doc-qt5/code/class_Circuit.html
        if not circuit:
            circuits = [c.name for c in netlist.each_circuit()]
            raise Exception(f"Expected circuit called {self.pex_context.top_cell.name} in extracted netlist, "
                            f"only available circuits are: {circuits}")

        diffusion_regions: List[kdb.Region] = []

        for net in circuit.each_net():
            # https://www.klayout.de/doc-qt5/code/class_Net.html
            debug(f"Net name={net.name}, expanded_name={net.expanded_name()}, pin_count={net.pin_count()}, "
                  f"is_floating={net.is_floating()}, is_passive={net.is_passive()}, "
                  f"terminals={list(map(lambda t: format_terminal(t), net.each_terminal()))}")

            net_name = net.expanded_name()

            for metal_layer in self.tech_info.process_metal_layers:
                metal_layer_name = metal_layer.name
                metal_layer = metal_layer.metal_layer

                metal_z_bottom = metal_layer.height
                metal_z_top = metal_z_bottom + metal_layer.thickness

                shapes = self.shapes_of_net(layer_name=metal_layer_name, net=net)
                if shapes:
                    if shapes.count() >= 1:
                        info(f"Conductor {net_name}, metal {metal_layer_name}, "
                             f"z={metal_layer.height}, height={metal_layer.thickness}")
                        model_builder.add_conductor(net_name=net_name,
                                                    layer=shapes,
                                                    z=metal_layer.height,
                                                    height=metal_layer.thickness)

                if metal_layer.HasField('contact_above'):
                    contact = metal_layer.contact_above
                    shapes = self.shapes_of_net(layer_name=contact.name, net=net)
                    if shapes and not shapes.is_empty():
                        info(f"Conductor {net_name}, via {contact.name}, "
                             f"z={metal_z_top}, height={contact.thickness}")
                        model_builder.add_conductor(net_name=net_name,
                                                    layer=shapes,
                                                    z=metal_z_top,
                                                    height=contact.thickness)

                # diel_above = self.tech_info.process_stack_layer_by_name.get(metal_layer.reference_above, None)
                # if diel_above:
                #     #model_builder.add_dielectric(material_name=metal_layer.reference_above,
                #     #                             layer=kdb.Region().)
                #     pass
                # TODO: add stuff

            # DIFF / TAP
            for diffusion_layer in self.tech_info.process_diffusion_layers:
                diffusion_layer_name = diffusion_layer.name
                diffusion_layer = diffusion_layer.diffusion_layer
                shapes = self.shapes_of_net(layer_name=diffusion_layer_name, net=net)
                if shapes and not shapes.is_empty():
                    diffusion_regions.append(shapes)
                    info(f"Diffusion {net_name}, layer {diffusion_layer_name}, "
                         f"z={0}, height={0.1}")
                    model_builder.add_conductor(net_name=net_name,
                                                layer=shapes,
                                                z=0,  # TODO
                                                height=0.1)  # TODO: diffusion_layer.height

                contact = diffusion_layer.contact_above
                shapes = self.shapes_of_net(layer_name=contact.name, net=net)
                if shapes and not shapes.is_empty():
                    info(f"Diffusion {net_name}, contact {contact.name}, "
                         f"z={0}, height={contact.thickness}")
                    model_builder.add_conductor(net_name=net_name,
                                                layer=shapes,
                                                z=0.0,
                                                height=contact.thickness)

        enlarged_top_cell_bbox = self.top_cell_bbox().enlarged(math.floor(8 / self.dbu))  # 8µm fringe halo

        #
        # global substrate block below everything. independent of nets!
        #

        substrate_layer = self.tech_info.process_substrate_layer.substrate_layer
        substrate_region = kdb.Region()

        substrate_block = enlarged_top_cell_bbox.dup()
        substrate_region.insert(substrate_block)

        diffusion_margin = math.floor(1 / self.dbu)  # 1 µm
        for d in diffusion_regions:
            substrate_region -= d.sized(diffusion_margin)
        info(f"Substrate VSUBS, "
             f"z={0 - substrate_layer.height - substrate_layer.thickness}, height={substrate_layer.thickness}")
        model_builder.add_conductor(net_name="VSUBS",
                                    layer=substrate_region,
                                    z=0 - substrate_layer.height - substrate_layer.thickness,
                                    height=substrate_layer.thickness)

        #
        # add dielectrics
        #

        fox_region = kdb.Region()
        fox_block = enlarged_top_cell_bbox.dup()
        fox_region.insert(fox_block)

        # field oxide goes from substrate/diff/well up to below the gate-poly
        gate_poly_height = self.tech_info.gate_poly_layer.metal_layer.height
        fox_z = 0
        fox_height = gate_poly_height - fox_z
        info(f"Simple dielectric (field oxide) {fox_layer.name}: "
             f"z={fox_z}, height={fox_height}")
        model_builder.add_dielectric(material_name=fox_layer.name,
                                     layer=fox_region,
                                     z=fox_z,
                                     height=fox_height)

        for metal_layer in self.tech_info.process_metal_layers:
            metal_layer_name = metal_layer.name
            metal_layer = metal_layer.metal_layer

            metal_z_bottom = metal_layer.height

            extracted_shapes = self.shapes_of_layer(layer_name=metal_layer_name)

            sidewall_region: Optional[kdb.Region] = None
            sidewall_height = 0

            no_metal_region: Optional[kdb.Region] = None
            no_metal_height = 0

            #
            # add sidewall dielectrics
            #
            if extracted_shapes:
                sidewall_height = 0
                sidewall_region = extracted_shapes
                sidewallee = metal_layer_name

                while True:
                    sidewall = self.tech_info.sidewall_dielectric_layer(sidewallee)
                    if not sidewall:
                        break
                    match sidewall.layer_type:
                        case process_stack_pb2.ProcessStackInfo.LAYER_TYPE_SIDEWALL_DIELECTRIC:
                            d = math.floor(sidewall.sidewall_dielectric_layer.width_outside_sidewall / self.dbu)
                            sidewall_region = sidewall_region.sized(d)
                            h_delta = sidewall.sidewall_dielectric_layer.height_above_metal or metal_layer.thickness
                            # if h_delta == 0:
                            #     h_delta = metal_layer.thickness
                            sidewall_height += h_delta
                            info(f"Sidewall dielectric {sidewall.name}: z={metal_layer.height}, height={sidewall_height}")
                            model_builder.add_dielectric(material_name=sidewall.name,
                                                         layer=sidewall_region,
                                                         z=metal_layer.height,
                                                         height=sidewall_height)

                        case process_stack_pb2.ProcessStackInfo.LAYER_TYPE_CONFORMAL_DIELECTRIC:
                            conf_diel = sidewall.conformal_dielectric_layer
                            d = math.floor(conf_diel.thickness_sidewall / self.dbu)
                            sidewall_region = sidewall_region.sized(d)
                            h_delta = metal_layer.thickness + conf_diel.thickness_over_metal
                            sidewall_height += h_delta
                            info(f"Conformal dielectric (sidewall) {sidewall.name}: "
                                 f"z={metal_layer.height}, height={sidewall_height}")
                            model_builder.add_dielectric(material_name=sidewall.name,
                                                         layer=sidewall_region,
                                                         z=metal_layer.height,
                                                         height=sidewall_height)
                            if conf_diel.thickness_where_no_metal > 0.0:
                                no_metal_block = enlarged_top_cell_bbox.dup()
                                no_metal_region = kdb.Region()
                                no_metal_region.insert(no_metal_block)
                                no_metal_region -= sidewall_region
                                no_metal_height = conf_diel.thickness_where_no_metal
                                info(f"Conformal dielectric (where no metal) {sidewall.name}: "
                                     f"z={metal_layer.height}, height={no_metal_height}")
                                model_builder.add_dielectric(material_name=sidewall.name,
                                                             layer=no_metal_region,
                                                             z=metal_layer.height,
                                                             height=no_metal_height)

                    sidewallee = sidewall.name

            #
            # add simple dielectric
            #
            simple_dielectric, diel_height = self.tech_info.simple_dielectric_above_metal(metal_layer_name)
            if simple_dielectric:
                diel_block = enlarged_top_cell_bbox.dup()
                diel_region = kdb.Region()
                diel_region.insert(diel_block)
                if sidewall_region:
                    assert sidewall_height >= 0.0
                    diel_region -= sidewall_region
                    info(f"Simple dielectric (sidewall) {simple_dielectric.name}: "
                         f"z={metal_z_bottom + sidewall_height}, height={diel_height - sidewall_height}")
                    model_builder.add_dielectric(material_name=simple_dielectric.name,
                                                 layer=sidewall_region,
                                                 z=metal_z_bottom + sidewall_height,
                                                 height=diel_height - sidewall_height)
                if no_metal_region:
                    info(f"Simple dielectric (no metal) {simple_dielectric.name}: "
                         f"z={metal_z_bottom + no_metal_height}, height={diel_height - no_metal_height}")
                    model_builder.add_dielectric(material_name=simple_dielectric.name,
                                                 layer=diel_region,
                                                 z=metal_z_bottom + no_metal_height,
                                                 height=diel_height - no_metal_height)
                else:
                    info(f"Simple dielectric {simple_dielectric.name}: "
                         f"z={metal_z_bottom}, height={diel_height}")
                    model_builder.add_dielectric(material_name=simple_dielectric.name,
                                                 layer=diel_region,
                                                 z=metal_z_bottom,
                                                 height=diel_height)

        gen = model_builder.generate()
        return gen



#
# --------------------------------------------------------------------------------
# SPDX-FileCopyrightText: 2024 Martin Jan Köhler and Harald Pretl
# Johannes Kepler University, Institute for Integrated Circuits.
#
# This file is part of KPEX 
# (see https://github.com/martinjankoehler/klayout-pex).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# SPDX-License-Identifier: GPL-3.0-or-later
# --------------------------------------------------------------------------------
#
from typing import *

import klayout.db as kdb

from ..log import (
    info,
)


class NetlistReducer:
    @staticmethod
    def reduce(netlist: kdb.Netlist,
               top_cell_name: str,
               cap_threshold: float = 0.05e-15) -> kdb.Netlist:
        reduced_netlist: kdb.Netlist = netlist.dup()
        reduced_netlist.combine_devices()  # merge C/R

        top_circuit: kdb.Circuit = reduced_netlist.circuit_by_name(top_cell_name)

        devices_to_remove: List[kdb.Device] = []

        for d in top_circuit.each_device():
            d: kdb.Device
            dc = d.device_class()
            if isinstance(dc, kdb.DeviceClassCapacitor):
                # net_a = d.net_for_terminal('A')
                # net_b = d.net_for_terminal('B')
                c_value = d.parameter('C')
                if c_value < cap_threshold:
                    devices_to_remove.append(d)

            elif isinstance(dc, kdb.DeviceClassResistor):
                # TODO
                pass

        for d in devices_to_remove:
            info(f"Removed device {d.name} {d.parameter('C')}")
            top_circuit.remove_device(d)

        return reduced_netlist
